"""Test using rgb mesh rendering."""


import numpy as np

from deodr.differentiable_renderer import Scene2D


def test_upper_left_pixel_center_coordinates() -> None:
    """Testing  that pixel center coordinates:

    pixel centers are at integer coordinates when integer_pixel_centers=True with
        upper left at (0, 0)
        upper right at (width - 1, 0)
        lower left at (0, height - 1)
        lower right at  (width - 1, height - 1)

    pixel centers are at half integer coordinates when integer_pixel_centers=False with
        upper left at (0.5, 0.5)
        upper right at (width - 0.5, 0.5)
        lower left at (0.5, height - 0.5)
        lower right at  (width - 0.5, height - 0.5)
    """
    height = 4
    width = 3
    integer_points_coordinates = [
        (0, 0),  # upper left
        (width - 1, 0),  # upper right,
        (0, height - 1),  # lower left,
        (width - 1, height - 1),  # lower right
    ]
    eps = 0.001

    clockwise = True
    for integer_pixel_centers in [False, True]:
        if integer_pixel_centers:
            points_coordinates = [
                (0.0, 0.0),  # upper left
                (width - 1.0, 0.0),  # upper right,
                (0, height - 1.0),  # lower left,
                (width - 1.0, height - 1.0),  # lower right
            ]
        else:
            points_coordinates = [
                (0.5, 0.5),  # upper left
                (width - 0.5, 0.5),  # upper right,
                (0.5, height - 0.5),  # lower left,
                (width - 0.5, height - 0.5),  # lower right
            ]

        depths = np.array([1, 1, 1])
        shade = np.array([0, 0, 0])
        shade = np.array([1, 1, 1])
        faces_uv = np.array([[0, 2, 1]], dtype=np.uint32)
        uv = np.zeros((3, 2), dtype=bool)
        textured = np.array([0], dtype=bool)
        shaded = np.array([0], dtype=bool)
        colors = np.array([[1], [1], [1]])
        edgeflags = np.zeros((1, 3), dtype=bool)
        faces = np.array([[0, 2, 1]], dtype=np.uint32)

        texture = np.ones((2, 2, 1))
        background_color = np.array([0])

        for integer_point_coordinates, point_coordinates in zip(
            integer_points_coordinates, points_coordinates
        ):
            ij = np.array([[-eps, -eps], [-eps, eps], [eps, -eps]]) + np.array(
                point_coordinates
            )

            scene2D = Scene2D(
                ij=ij,
                faces=faces,
                faces_uv=faces_uv,
                uv=uv,
                texture=texture,
                height=height,
                width=width,
                nb_colors=1,
                background_image=None,
                background_color=background_color,
                depths=depths,
                textured=textured,
                shade=shade,
                colors=colors,
                shaded=shaded,
                edgeflags=edgeflags,
                strict_edge=False,
                perspective_correct=True,
                clockwise=clockwise,
                integer_pixel_centers=integer_pixel_centers,
            )

            image, _ = scene2D.render(sigma=0)

            expected_image = np.zeros((height, width, 1))
            expected_image[
                integer_point_coordinates[1], integer_point_coordinates[0], 0
            ] = 1
            assert np.allclose(expected_image, image)


if __name__ == "__main__":
    test_upper_left_pixel_center_coordinates()



"""Implementation of triangulated meshes."""
from typing import Any, Dict, Optional, Tuple

import numpy as np
from scipy import sparse

from .tools import cross_backward, normalize, normalize_backward

try:
    from trimesh.base import Trimesh
except ImportError:
    Trimesh = None


class TriMeshAdjacencies:
    """Class that stores sparse adjacency matrices and methods that use these matrices.
    Unlike the TriMesh class there are no vertices stored in this class.
    """

    def __init__(
        self,
        faces: np.ndarray,
        clockwise: bool = False,
        nb_vertices: Optional[int] = None,
    ):
        faces = np.array(faces)
        assert faces.ndim == 2
        assert faces.shape[1] == 3

        self.faces = faces

        self.nb_faces = int(faces.shape[0])
        if nb_vertices is None:
            nb_vertices = int(np.max(faces.flat)) + 1
        self.nb_vertices = nb_vertices
        i = self.faces.flatten()
        j = np.tile(np.arange(self.nb_faces)[:, None], [1, 3]).flatten()
        v = np.ones((self.nb_faces, 3)).flatten()
        self._vertices_faces = sparse.coo_matrix(
            (v, (i, j)), shape=(self.nb_vertices, self.nb_faces)
        )
        id_faces = np.hstack(
            (
                np.arange(self.nb_faces),
                np.arange(self.nb_faces),
                np.arange(self.nb_faces),
            )
        )
        self.clockwise = clockwise
        edges = np.vstack(
            (self.faces[:, [0, 1]], self.faces[:, [1, 2]], self.faces[:, [2, 0]])
        )

        id_edge_tmp, edge_increase = self.id_edge(edges)

        _, id_edge, unique_counts = np.unique(
            id_edge_tmp, return_inverse=True, return_counts=True
        )

        self.nb_edges = np.max(id_edge) + 1
        self.edges = np.zeros((self.nb_edges, 2), dtype=np.uint32)
        self.edges[id_edge] = edges

        nb_inc = np.zeros((self.nb_edges))
        np.add.at(nb_inc, id_edge, edge_increase)
        nb_dec = np.zeros((self.nb_edges))
        np.add.at(nb_dec, id_edge, ~edge_increase)
        self.is_manifold = (
            np.all(unique_counts <= 2) and np.all(nb_inc <= 1) and np.all(nb_dec <= 1)
        )
        self.is_closed = self.is_manifold and np.all(unique_counts == 2)

        self.edges_vertices_ones = sparse.coo_matrix(
            (
                np.ones((2 * len(id_edge))),
                (np.tile(id_edge[:, None], (1, 2)).flatten(), edges.flatten()),
            ),
            shape=(self.nb_edges, self.nb_vertices),
        )

        self.edges_faces_ones = sparse.coo_matrix(
            (np.ones((len(id_edge))), (id_edge, id_faces)),
            shape=(self.nb_edges, self.nb_faces),
        )
        v = np.hstack(
            (
                np.full((self.nb_faces), 0),
                np.full((self.nb_faces), 1),
                np.full((self.nb_faces), 2),
            )
        )
        self.faces_edges = sparse.coo_matrix(
            (id_edge, (id_faces, v)), shape=(self.nb_faces, 3)
        ).todense()
        self.adjacency_vertices = (
            (self._vertices_faces * self._vertices_faces.T) > 0
        ) - sparse.eye(self.nb_vertices)

        self.degree_v_f = self._vertices_faces.dot(np.ones((self.nb_faces)))

        self.degree_v_e = self.adjacency_vertices.dot(
            np.ones((self.nb_vertices))
        )  # degree_v_e(i)=j means that the vertex i appears in j edges
        self.laplacian = (
            sparse.diags([self.degree_v_e], [0], (self.nb_vertices, self.nb_vertices))
            - self.adjacency_vertices
        )
        self.hasBoundaries = np.any(np.sum(self.edges_faces_ones, axis=1) == 1)
        assert np.all(self.laplacian * np.ones((self.nb_vertices)) == 0)
        self.store_backward: Dict[str, Any] = {}

    def boundary_edges(self) -> np.ndarray:
        is_boundary_edge = np.array(np.sum(self.edges_faces_ones, axis=1) == 1).squeeze(
            axis=1
        )
        return self.edges[is_boundary_edge, :]

    def id_edge(self, idv: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        assert idv.ndim == 2
        assert idv.shape[1] == 2
        return (
            np.maximum(idv[:, 0], idv[:, 1]).astype(np.uint64)
            + np.minimum(idv[:, 0], idv[:, 1]).astype(np.uint64) * self.nb_vertices,
            idv[:, 0] < idv[:, 1],
        )

    def compute_face_normals(self, vertices: np.ndarray) -> np.ndarray:
        assert vertices.ndim == 2
        assert vertices.shape[1] == 3
        triangles = vertices[self.faces, :]
        u = triangles[:, 1, :] - triangles[:, 0, :]
        v = triangles[:, 2, :] - triangles[:, 0, :]
        n = -np.cross(u, v) if self.clockwise else np.cross(u, v)
        normals = normalize(n, axis=1)
        self.store_backward["compute_face_normals"] = (u, v, n)
        return normals

    def compute_face_normals_backward(self, normals_b: np.ndarray) -> np.ndarray:
        assert normals_b.ndim == 2
        assert normals_b.shape[1] == 3
        u, v, n = self.store_backward["compute_face_normals"]
        n_b = normalize_backward(n, normals_b, axis=1)
        if self.clockwise:
            u_b, v_b = cross_backward(u, v, -n_b)
        else:
            u_b, v_b = cross_backward(u, v, n_b)
        triangles_b = np.stack((-u_b - v_b, u_b, v_b), axis=1)
        vertices_b = np.zeros((self.nb_vertices, 3))
        np.add.at(vertices_b, self.faces, triangles_b)
        return vertices_b

    def compute_vertex_normals(self, face_normals: np.ndarray) -> np.ndarray:
        assert face_normals.ndim == 2
        assert face_normals.shape[1] == 3
        n = self._vertices_faces * face_normals
        normals = normalize(n, axis=1)
        self.store_backward["compute_vertex_normals"] = n
        return normals

    def compute_vertex_normals_backward(self, normals_b: np.ndarray) -> np.ndarray:
        assert normals_b.ndim == 2
        assert normals_b.shape[1] == 3
        n = self.store_backward["compute_vertex_normals"]
        n_b = normalize_backward(n, normals_b, axis=1)
        return self._vertices_faces.T * n_b

    def edge_on_silhouette(self, vertices_2d: np.ndarray) -> np.ndarray:
        """Compute the a boolean for each of edges of each face that is true if
        and only if the edge is one the silhouette of the mesh given a view point
        """
        assert vertices_2d.ndim == 2
        assert vertices_2d.shape[1] == 2
        triangles = vertices_2d[self.faces, :]
        u = triangles[:, 1, :] - triangles[:, 0, :]
        v = triangles[:, 2, :] - triangles[:, 0, :]
        face_visible = np.cross(u, v) > 0 if self.clockwise else np.cross(u, v) < 0
        edge_bool = (self.edges_faces_ones * face_visible) == 1
        return edge_bool[self.faces_edges]


class TriMesh:
    """Class that implements a triangulated mesh."""

    def __init__(
        self,
        faces: np.ndarray,
        vertices: np.ndarray,
        clockwise: bool = False,
        compute_adjacencies: bool = True,
    ):
        faces = np.array(faces)
        assert np.issubdtype(faces.dtype, np.integer)
        assert faces.ndim == 2
        assert faces.shape[1] == 3
        assert np.all(faces >= 0)

        self._faces = faces
        self.nb_vertices = int(np.max(faces)) + 1
        self.nb_faces = int(faces.shape[0])

        self._face_normals: Optional[np.ndarray] = None
        self._vertex_normals: Optional[np.ndarray] = None
        self.clockwise = clockwise

        self._vertices_b = np.zeros((self.nb_vertices, 3))

        self.set_vertices(vertices)
        if compute_adjacencies:
            self.compute_adjacencies()

    def compute_adjacencies(self) -> None:
        self._adjacencies = TriMeshAdjacencies(
            self.faces, self.clockwise, nb_vertices=self.nb_vertices
        )

        if self._adjacencies.is_closed:
            self.check_orientation()

    @property
    def vertices(self) -> np.ndarray:
        return self._vertices

    @property
    def faces(self) -> np.ndarray:
        return self._faces

    @property
    def adjacencies(self) -> TriMeshAdjacencies:
        if self._adjacencies is None:
            self.compute_adjacencies()
        return self._adjacencies

    def set_vertices(self, vertices: np.ndarray) -> None:
        assert vertices.ndim == 2
        assert vertices.shape[1] == 3
        self._vertices = vertices
        self._face_normals = None
        self._vertex_normals = None
        self._vertices_b = np.zeros((self.nb_vertices, 3))

    def compute_volume(self) -> float:
        """Compute the volume enclosed by the triangulated surface. It assumes the
        surfaces is a closed manifold. This is done by summing the volumes of the
        simplices formed by joining the origin and the vertices of each triangle.
        """
        return (
            (1 if self.clockwise else -1)
            * np.sum(
                np.linalg.det(
                    np.dstack(
                        (
                            self.vertices[self._faces[:, 0]],
                            self.vertices[self._faces[:, 1]],
                            self.vertices[self._faces[:, 2]],
                        )
                    )
                )
            )
            / 6
        )

    def check_orientation(self) -> None:
        """Check the mesh faces are properly oriented for the normals to point
        outward.
        """
        if self.compute_volume() > 0:
            raise (
                BaseException(
                    "The volume within the surface is negative. It seems that you faces"
                    "are not oriented correctly according to the clockwise flag"
                )
            )

    def compute_face_normals(self) -> None:
        self._face_normals = self.adjacencies.compute_face_normals(self.vertices)

    @property
    def face_normals(self) -> np.ndarray:
        """Return the face normals.

        face normals evaluation is done in a lazy manner.
        """
        if self._face_normals is None:
            self.compute_face_normals()
        assert self._face_normals is not None
        return self._face_normals

    def compute_vertex_normals(self) -> None:
        self._vertex_normals = self.adjacencies.compute_vertex_normals(
            self.face_normals
        )

    @property
    def vertex_normals(self) -> np.ndarray:
        """Return the vertices normals.

        face normals evaluation is done in a lazy manner.
        """

        if self._vertex_normals is None:
            self.compute_vertex_normals()
        assert self._vertex_normals is not None
        return self._vertex_normals

    def compute_vertex_normals_backward(self, vertex_normals_b: np.ndarray) -> None:
        self._face_normals_b = self.adjacencies.compute_vertex_normals_backward(
            vertex_normals_b
        )
        self._vertices_b += self.adjacencies.compute_face_normals_backward(
            self._face_normals_b
        )

    def edge_on_silhouette(self, points_2d: np.ndarray) -> np.ndarray:
        """Compute the a boolean for each of edges that is true if and only if
        the edge is one the silhouette of the mesh.
        """
        assert self.adjacencies.is_manifold
        return self.adjacencies.edge_on_silhouette(points_2d)


class ColoredTriMesh(TriMesh):
    """Class that implements a colored triangulated mesh."""

    def __init__(
        self,
        faces: np.ndarray,
        vertices: np.ndarray,
        clockwise: bool = False,
        faces_uv: Optional[np.ndarray] = None,
        uv: Optional[np.ndarray] = None,
        texture: Optional[np.ndarray] = None,
        colors: Optional[np.ndarray] = None,
        nb_colors: Optional[int] = None,
        compute_adjacencies: bool = True,
    ):
        super(ColoredTriMesh, self).__init__(
            faces,
            vertices=vertices,
            clockwise=clockwise,
            compute_adjacencies=compute_adjacencies,
        )
        self.faces_uv = faces_uv
        self.uv = uv

        self.texture = texture
        self.vertices_colors = colors
        self.textured = self.texture is not None
        self.nb_colors = nb_colors
        if nb_colors is None:
            if texture is None:
                assert (
                    colors is not None
                ), "You need to provide at least on among nb_colors, texture or colors"
                self.nb_colors = colors.shape[1]
            else:
                assert (
                    texture is not None
                ), "You need to provide at least on among nb_colors, texture or colors"
                self.nb_colors = texture.shape[2]

        self.vertices_colors_b: Optional[np.ndarray] = None

    def set_vertices_colors(self, colors: np.ndarray) -> None:
        self.vertices_colors = colors

    def plot_uv_map(self, ax: Any) -> None:
        assert self.uv is not None, "You need to provide a uv to display the uv map"
        if self.texture is not None:
            ax.imshow(self.texture)
        ax.triplot(self.uv[:, 0], self.uv[:, 1], self.faces_uv)

    def plot(self, ax: Any) -> None:
        assert self.vertices is not None, "You need to provide vertices first"
        x, y, z = self.vertices.T
        u, v, w = self.vertex_normals.T
        ax.plot_trisurf(
            self.vertices[:, 0],
            self.vertices[:, 1],
            Z=self.vertices[:, 2],
            triangles=self.faces,
        )
        ax.quiver(x, y, z, u, v, w, length=0.03, normalize=True, color=[0, 1, 0])

    def subdivise(self, n_iter: int) -> "ColoredTriMesh":
        """loop subdivision.

        https://graphics.stanford.edu/~mdfisher/subdivision.html"""
        return loop_subdivision(self, n_iter)

    @staticmethod
    def from_trimesh(
        mesh: Trimesh, compute_adjacencies: bool = True
    ) -> "ColoredTriMesh":  # inspired from pyrender
        """Get the vertex colors, texture coordinates, and material properties
        from a :class:`~trimesh.base.Trimesh`.
        """
        colors = None
        uv = None
        texture: Optional[np.ndarray] = None

        # If the trimesh visual is undefined, return none for both

        # Process vertex colors
        if mesh.visual.kind == "vertex":
            colors = mesh.visual.vertex_colors.copy()
            if colors.ndim == 2 and colors.shape[1] == 4:
                colors = colors[:, :3]
            colors = colors.astype(np.float64) / 255

        # Process face colors
        elif mesh.visual.kind == "face":
            raise BaseException(
                "not supported yet, will need antialiasing at the seams"
            )

        # Process texture colors
        elif mesh.visual.kind == "texture":
            # Configure UV coordinates
            if mesh.visual.uv is not None:
                texture = np.array(mesh.visual.material.image) / 255
                texture.setflags(write=False)

                if texture.shape[2] == 4:
                    texture = texture[:, :, :3]  # removing alpha channel
                assert texture is not None  # helping mypy

                uv = (
                    np.column_stack(
                        (
                            (mesh.visual.uv[:, 0]) * texture.shape[1],
                            (1 - mesh.visual.uv[:, 1]) * texture.shape[0],
                        )
                    )
                    - 0.5
                )

        # merge identical 3D vertices even if their uv are different to keep surface
        # manifold. Trimesh seems to split vertices that have different uvs (using
        # unmerge_faces texture.py), making the surface not watertight, while there
        # were only seems in the texture.

        vertices, return_index, inv_ids = np.unique(
            mesh.vertices, axis=0, return_index=True, return_inverse=True
        )
        faces = inv_ids[mesh.faces].astype(np.uint32)
        if colors is not None:
            colors2 = colors[return_index, :]
            if np.any(colors != colors2[inv_ids, :]):
                raise (
                    BaseException(
                        "vertices at the same 3D location should have the same color"
                        "for the rendering to be differentiable"
                    )
                )
        else:
            colors2 = None

        return ColoredTriMesh(
            faces,
            vertices,
            clockwise=False,
            faces_uv=np.array(mesh.faces),
            uv=uv,
            texture=texture,
            colors=colors2,
            compute_adjacencies=compute_adjacencies,
        )

    def to_trimesh(self) -> Trimesh:
        # lazy modules loading
        import PIL
        import trimesh

        # largely inspired from trimesh's load_obj function

        if self.vertices_colors is not None:
            raise BaseException(
                "Conversion to trimesh with per vertex color not support yet"
            )

        v = self.vertices
        faces = self.faces
        faces_tex = self.faces_uv

        assert self.uv is not None, "Only mesh with texture supported."
        assert self.texture is not None, "Only mesh with texture supported."

        vt = np.column_stack(
            (
                (self.uv[:, 0] + 0.5) / self.texture.shape[1],
                1 - ((self.uv[:, 1] + 0.5) / self.texture.shape[0]),
            )
        )
        new_faces, mask_v, mask_vt = trimesh.visual.texture.unmerge_faces(
            faces, faces_tex
        )
        assert np.allclose(v[faces], v[mask_v][new_faces])
        assert new_faces.max() < len(v[mask_v])

        new_vertices = v[mask_v].copy()
        uv = vt[mask_vt].copy()

        texture_uint8 = np.clip(self.texture * 255, 0, 255).astype(np.uint8)
        if texture_uint8.shape[2] == 1:
            texture_uint8 = texture_uint8.squeeze(axis=2)
        texture_pil = PIL.Image.fromarray(texture_uint8)
        material = trimesh.visual.material.SimpleMaterial(image=texture_pil)
        visual = trimesh.visual.texture.TextureVisuals(uv=uv, material=material)

        return trimesh.Trimesh(vertices=new_vertices, faces=new_faces, visual=visual)

    @staticmethod
    def load(filename: str) -> "ColoredTriMesh":
        import trimesh

        mesh_trimesh = trimesh.load(filename)
        return ColoredTriMesh.from_trimesh(mesh_trimesh)


def loop_subdivision(mesh: ColoredTriMesh, n_iter: int = 1) -> ColoredTriMesh:
    """Loop subdivision.

    https://graphics.stanford.edu/~mdfisher/subdivision.html"""

    if n_iter == 0:
        return mesh

    if n_iter > 1:
        mesh = loop_subdivision(mesh, n_iter - 1)

    edge_mid_points = (
        mesh.adjacencies.edges_faces_ones
        * (mesh.adjacencies._vertices_faces.T * mesh.vertices)
        / 8
    ) + (1 / 8) * np.sum(mesh.vertices[mesh.adjacencies.edges, :], axis=1)

    # edge_mid_points = 0.5 * np.sum(self.vertices[self.adjacencies.edges, :], axis=1)

    beta = (3 / 8) * (1 / mesh.adjacencies.degree_v_e)
    moved_points = (
        beta[:, None] * (mesh.adjacencies.adjacency_vertices * mesh.vertices)
        + (5 / 8) * mesh.vertices
    )
    # moved_points = self.vertices

    new_vertices = np.vstack((moved_points, edge_mid_points))
    faces1 = np.column_stack(
        (
            mesh.faces[:, 0],
            mesh.adjacencies.faces_edges[:, 0] + mesh.nb_vertices,
            mesh.adjacencies.faces_edges[:, 2] + mesh.nb_vertices,
        )
    )
    faces2 = np.column_stack(
        (
            mesh.faces[:, 1],
            mesh.adjacencies.faces_edges[:, 1] + mesh.nb_vertices,
            mesh.adjacencies.faces_edges[:, 0] + mesh.nb_vertices,
        )
    )
    faces3 = np.column_stack(
        (
            mesh.faces[:, 2],
            mesh.adjacencies.faces_edges[:, 2] + mesh.nb_vertices,
            mesh.adjacencies.faces_edges[:, 1] + mesh.nb_vertices,
        )
    )
    faces4 = np.column_stack(
        (
            mesh.adjacencies.faces_edges[:, 0] + mesh.nb_vertices,
            mesh.adjacencies.faces_edges[:, 1] + mesh.nb_vertices,
            mesh.adjacencies.faces_edges[:, 2] + mesh.nb_vertices,
        )
    )
    new_faces = np.row_stack((faces1, faces2, faces3, faces4))
    if mesh.uv is not None:
        raise BaseException("Textured mesh not supported yet in subdivision.")
    if mesh.vertices_colors is not None:
        edge_mid_points_colors = np.mean(
            mesh.vertices_colors[mesh.adjacencies.edges, :], axis=1
        )
        new_colors = np.vstack((mesh.vertices_colors, edge_mid_points_colors))
    else:
        new_colors = None
    return ColoredTriMesh(
        vertices=new_vertices,
        faces=new_faces,
        colors=new_colors,
        nb_colors=mesh.nb_colors,
    )



# distutils: language = c++
from libcpp cimport bool

import cython

# import both numpy and the Cython declarations for numpy
import numpy as np

cimport numpy as np

cdef extern from "../C++/DifferentiableRenderer.h":
	ctypedef struct Scene:
		unsigned int* faces;
		unsigned int* faces_uv;
		double* depths
		double* uv
		double* ij
		double* shade
		double* colors
		bool* edgeflags
		bool* textured
		bool* shaded
		int     nb_triangles
		int nb_vertices;
		bool clockwise;
		bool backface_culling;
		int nb_uv;
		int     height
		int     width
		int     nb_colors
		double* texture
		int  texture_height
		int  texture_width
		double* background_image
		double* background_color
		double* uv_b
		double* ij_b
		double* shade_b
		double* colors_b
		double* texture_b
		bool strict_edge
		bool perspective_correct
		bool integer_pixel_centers
	void renderScene(Scene scene,double* image,double* z_buffer,double sigma,bool antialiase_error ,double* obs,double*  err_buffer)
	void renderScene_B(Scene scene,double* image,double* z_buffer,double* image_b,double sigma,bool antialiase_error ,double* obs,double*  err_buffer, double* err_buffer_b)


@cython.boundscheck(False)
@cython.wraparound(False)
def renderSceneCpp(scene, 
		double sigma,
		np.ndarray[double,ndim = 3,mode = "c"] image, 
		np.ndarray[double,ndim = 2,mode = "c"] z_buffer,
		bool antialiase_error  = 0,
		np.ndarray[double,ndim = 3,mode = "c"] obs = None,
		np.ndarray[double,ndim = 2,mode = "c"] err_buffer = None, 
		bool check_valid = 1):
 
	cdef Scene scene_c 

	if check_valid:
		assert (not(image is None))
		assert (not(z_buffer is None))

	height  =  image.shape[0]
	width  =  image.shape[1]
	nb_colors  =  image.shape[2]
		
	nb_triangles  =  scene.faces.shape[0]
	assert(nb_triangles  ==  scene.faces_uv.shape[0])
	nb_vertices  =  scene.depths.shape[0]
	nb_vertices_uv  =  scene.uv.shape[0]

	if check_valid:	
		assert(scene.faces.dtype  ==  np.uint32)
		assert(np.all(scene.faces < nb_vertices))
		assert(np.all(scene.faces_uv < nb_vertices_uv))
				
		assert(scene.colors.ndim  ==  2)
		assert(scene.uv.ndim  ==  2)
		assert(scene.ij.ndim  ==  2)
		assert(scene.shade.ndim  ==  1)
		assert(scene.edgeflags.ndim  ==  2)
		assert(scene.textured.ndim  ==  1)
		assert(scene.shaded.ndim  ==  1)		
		assert(scene.uv.shape[1]  ==  2)	
		assert(scene.ij.shape[0]  ==  nb_vertices)
		assert(scene.ij.shape[1]  ==  2)
		assert(scene.shade.shape[0]  ==  nb_vertices)
		assert(scene.colors.shape[0]  ==  nb_vertices)
		assert(scene.colors.shape[1]  ==  nb_colors)
		assert(scene.edgeflags.shape[0]  ==  nb_triangles)
		assert(scene.edgeflags.shape[1]  ==  3)
		assert(scene.textured.shape[0]  ==  nb_triangles)
		assert(scene.shaded.shape[0]  ==  nb_triangles)
		
		assert ((scene.background_image is not None) != (scene.background_color is not None))

		if scene.background_image is not None:
			assert scene.background_image.ndim == 3
			assert scene.background_image.shape[0] == height
			assert scene.background_image.shape[1] == width
			assert scene.background_image.shape[2] == nb_colors
		else:
			assert scene.background_color.shape[0] == nb_colors
		
		if scene.texture.size>0:
			assert(scene.texture.ndim  ==  3)
			assert(scene.texture.shape[0]>0)
			assert(scene.texture.shape[1]>0)
			assert(scene.texture.shape[2]  ==  nb_colors)
		
		assert z_buffer.shape[0]  ==  height 
		assert z_buffer.shape[1]  ==  width 

	scene_c.nb_colors = nb_colors
	cdef np.ndarray[np.uint32_t, mode = "c"] faces_c  =  np.ascontiguousarray(scene.faces.flatten(), dtype = np.uint32)
	cdef np.ndarray[np.uint32_t, mode = "c"] faces_uv_c  =  np.ascontiguousarray(scene.faces_uv.flatten(), dtype = np.uint32)		
	cdef np.ndarray[np.double_t, mode = "c"] depths_c  =  np.ascontiguousarray(scene.depths.flatten(), dtype = np.double)	
	cdef np.ndarray[np.double_t, mode = "c"] uv_c  =  np.ascontiguousarray(scene.uv.flatten(), dtype = np.double)
	cdef np.ndarray[np.double_t, mode = "c"] ij_c  =  np.ascontiguousarray(scene.ij.flatten(), dtype = np.double)
	cdef np.ndarray[np.double_t, mode = "c"] shade_c  =  np.ascontiguousarray(scene.shade.flatten(), dtype = np.double)
	cdef np.ndarray[np.double_t, mode = "c"] colors_c  =  np.ascontiguousarray(scene.colors.flatten(), dtype = np.double)
	cdef np.ndarray[np.uint8_t, mode = "c"] edgeflags_c  =  np.ascontiguousarray(scene.edgeflags.flatten(), dtype = np.uint8)
	cdef np.ndarray[np.uint8_t, mode = "c"] textured_c  =  np.ascontiguousarray(scene.textured.flatten(), dtype = np.uint8)
	cdef np.ndarray[np.uint8_t, mode = "c"] shaded_c  =  np.ascontiguousarray(scene.shaded.flatten(), dtype = np.uint8)
	cdef np.ndarray[np.double_t, mode = "c"] texture_c  =  np.ascontiguousarray(scene.texture.flatten(), dtype = np.double)
	cdef np.ndarray[np.double_t, mode = "c"] background_image_c;
	cdef np.ndarray[np.double_t, mode = "c"] background_color_c;

	if scene.background_image is not None:
		background_image_c =  np.ascontiguousarray(scene.background_image.flatten(), dtype = np.double)
	else:
		background_color_c =  np.ascontiguousarray(scene.background_color.flatten(), dtype = np.double)

	scene_c.height = <int> scene.height
	scene_c.width = <int> scene.width	
	scene_c.nb_triangles = nb_triangles
	scene_c.nb_vertices = nb_vertices
	scene_c.backface_culling = scene.backface_culling
	scene_c.clockwise = scene.clockwise
	scene_c.nb_uv = nb_vertices_uv
	scene_c.faces = <unsigned int*> faces_c.data
	scene_c.faces_uv = <unsigned int*> faces_uv_c.data
	scene_c.depths = <double*> depths_c.data
	scene_c.uv = <double*> uv_c.data
	scene_c.ij = <double*> ij_c.data
	scene_c.shade = <double*> shade_c.data
	scene_c.colors = <double*> colors_c.data
	scene_c.edgeflags = <bool*> edgeflags_c.data
	scene_c.textured = <bool*> textured_c.data
	scene_c.shaded = <bool*> shaded_c.data
	scene_c.texture = <double*> texture_c.data
	
	if scene.background_image is not None:	
		scene_c.background_image = <double*> background_image_c.data
		scene_c.background_color = NULL
		if scene_c.background_image  ==  NULL:
			raise BaseException('scene_c.background_image_c is NULL')

	else:
		scene_c.background_image = NULL
		scene_c.background_color = <double*> background_color_c.data
		if scene_c.background_color  ==  NULL:
			raise BaseException('scene_c.background_color is NULL')

	scene_c.texture_height = scene.texture.shape[0]
	scene_c.texture_width = scene.texture.shape[1]
	scene_c.strict_edge = scene.strict_edge
	scene_c.perspective_correct = scene.perspective_correct
	scene_c.integer_pixel_centers = scene.integer_pixel_centers


	cdef double* obs_ptr = NULL
	cdef double* err_buffer_ptr = NULL
	
	cdef double* image_ptr =  <double*> image.data
	cdef double* z_buffer_ptr =  <double*> z_buffer.data
	
	if image_ptr  ==  NULL:
		raise BaseException('image_ptr is NULL')
	if z_buffer_ptr  ==  NULL:
		raise BaseException('z_buffer_ptr is NULL')

	
	if antialiase_error:
		assert err_buffer.shape[0]  ==  height 
		assert err_buffer.shape[1]  ==  width
		assert obs.shape[0]  ==  height 
		assert obs.shape[1]  ==  width
		assert obs.shape[2]  ==  nb_colors 
		
		obs_ptr  =  <double*>obs.data
		err_buffer_ptr = <double*>err_buffer.data
		
		if err_buffer_ptr  ==  NULL:
			raise BaseException('err_buffer_ptr is NULL')
		if obs_ptr  ==  NULL:
			raise BaseException('obs_ptr is NULL')
	

	renderScene( scene_c,image_ptr, z_buffer_ptr, sigma, antialiase_error ,obs_ptr, err_buffer_ptr)
	
@cython.boundscheck(False)
@cython.wraparound(False)	
def renderSceneBCpp(scene, 
		double sigma,
		np.ndarray[double,ndim = 3,mode = "c"] image, 
		np.ndarray[double,ndim = 2,mode = "c"] z_buffer,
		np.ndarray[double,ndim = 3,mode = "c"] image_b = None,
		bool antialiase_error  = 0,
		np.ndarray[double,ndim = 3,mode = "c"] obs = None,
		np.ndarray[double,ndim = 2,mode = "c"] err_buffer = None,
		np.ndarray[double,ndim = 2,mode = "c"] err_buffer_b = None,
		bool check_valid=1):

	cdef Scene scene_c 

	if check_valid:
		assert (not(image is None))
		assert (not(z_buffer is None))
		
	height = image.shape[0]
	width  = image.shape[1]
	nb_colors = image.shape[2]
	nb_triangles  =  scene.faces.shape[0]
	
	if check_valid:
		assert(nb_colors  ==  scene.colors.shape[1])			
		assert z_buffer.shape[0]  ==  height 
		assert z_buffer.shape[1]  ==  width 	
		assert(nb_triangles  ==  scene.faces_uv.shape[0])

	nb_vertices  =  scene.depths.shape[0]
	nb_vertices_uv  =  scene.uv.shape[0]
	
	if check_valid:
		assert(scene.faces.dtype == np.uint32)
		assert(np.all(scene.faces<nb_vertices))
		assert(np.all(scene.faces_uv<nb_vertices_uv))
				
		assert(scene.colors.ndim  ==  2)
		assert(scene.uv.ndim  ==  2)
		assert(scene.ij.ndim  ==  2)
		assert(scene.shade.ndim  ==  1)
		assert(scene.edgeflags.ndim  ==  2)
		assert(scene.textured.ndim  ==  1)
		assert(scene.shaded.ndim  ==  1)		
		assert(scene.uv.shape[1]  ==  2)	
		assert(scene.ij.shape[0]  ==  nb_vertices)
		assert(scene.ij.shape[1]  ==  2)
		assert(scene.shade.shape[0]  ==  nb_vertices)
		assert(scene.colors.shape[0]  ==  nb_vertices)
		assert(scene.colors.shape[1]  ==  nb_colors)
		assert(scene.edgeflags.shape[0]  ==  nb_triangles)
		assert(scene.edgeflags.shape[1]  ==  3)
		assert(scene.textured.shape[0]  ==  nb_triangles)
		assert(scene.shaded.shape[0]  ==  nb_triangles)
		
		assert (scene.background_image is not None) != (scene.background_color is not None)

		if scene.background_image is not None:
			assert scene.background_image.ndim == 3
			assert scene.background_image.shape[0] == height
			assert scene.background_image.shape[1] == width
			assert scene.background_image.shape[2] == nb_colors
		else:
			assert scene.background_color.shape[0] == nb_colors
				
		assert(scene.uv_b.ndim  ==  2)
		assert(scene.ij_b.ndim  ==  2)
		assert(scene.shade_b.ndim  ==  1)
		assert(scene.edgeflags.ndim  ==  2)
		assert(scene.textured.ndim  ==  1)
		assert(scene.shaded.ndim  ==  1)	
		assert(scene.uv_b.shape[0]  ==  nb_vertices_uv)
		assert(scene.uv_b.shape[1]  ==  2)
		assert(scene.ij_b.shape[0]  ==  nb_vertices)
		assert(scene.ij_b.shape[1]  ==  2)
		assert(scene.shade_b.shape[0]  ==  nb_vertices)	
		assert(scene.colors_b.shape[0]  ==  nb_vertices)
		assert(scene.colors_b.shape[1]  ==  nb_colors)

					
		if scene.texture.size>0:
			assert(scene.texture.ndim  ==  3)
			assert(scene.texture_b.ndim  ==  3)	
			assert(scene.texture.shape[0]>0)
			assert(scene.texture.shape[1]>0)		
			assert(scene.texture.shape[0]  ==  scene.texture_b.shape[0])
			assert(scene.texture.shape[1]  ==  scene.texture_b.shape[1])
			assert(scene.texture.shape[2]  ==  nb_colors)	
			assert(scene.texture_b.shape[2]  ==  nb_colors)
			
	scene_c.nb_colors = nb_colors
	
	cdef np.ndarray[np.uint32_t, mode = "c"] faces_c  =  np.ascontiguousarray(scene.faces.flatten(), dtype = np.uint32)
	cdef np.ndarray[np.uint32_t, mode = "c"] faces_uv_c  =  np.ascontiguousarray(scene.faces_uv.flatten(), dtype = np.uint32)
	cdef np.ndarray[np.double_t, mode = "c"] depths_c =  np.ascontiguousarray(scene.depths.flatten(), dtype = np.double)	
	cdef np.ndarray[np.double_t, mode = "c"] uv_c =  np.ascontiguousarray(scene.uv.flatten(), dtype = np.double)
	cdef np.ndarray[np.double_t, mode = "c"] ij_c =  np.ascontiguousarray(scene.ij.flatten(), dtype = np.double)
	cdef np.ndarray[np.double_t, mode = "c"] uv_b_c =  np.ascontiguousarray(scene.uv_b.flatten(), dtype = np.double)
	cdef np.ndarray[np.double_t, mode = "c"] ij_b_c =  np.ascontiguousarray(scene.ij_b.flatten(), dtype = np.double)
	cdef np.ndarray[np.double_t, mode = "c"] shade_c =  np.ascontiguousarray(scene.shade.flatten(), dtype = np.double)
	cdef np.ndarray[np.double_t, mode = "c"] shade_b_c =  np.ascontiguousarray(scene.shade_b.flatten(), dtype = np.double)
	cdef np.ndarray[np.double_t, mode = "c"] colors_c =  np.ascontiguousarray(scene.colors.flatten(), dtype = np.double)
	cdef np.ndarray[np.double_t, mode = "c"] colors_b_c =  np.ascontiguousarray(scene.colors_b.flatten(), dtype = np.double)
	cdef np.ndarray[np.uint8_t, mode = "c"] edgeflags_c =  np.ascontiguousarray(scene.edgeflags.flatten(), dtype = np.uint8)
	cdef np.ndarray[np.uint8_t, mode = "c"] textured_c =  np.ascontiguousarray(scene.textured.flatten(), dtype = np.uint8)
	cdef np.ndarray[np.uint8_t, mode = "c"] shaded_c =  np.ascontiguousarray(scene.shaded.flatten(), dtype = np.uint8)
	cdef np.ndarray[np.double_t, mode = "c"] texture_c =  np.ascontiguousarray(scene.texture.flatten(), dtype = np.double)
	cdef np.ndarray[np.double_t, mode = "c"] texture_b_c =  np.ascontiguousarray(scene.texture_b.flatten(), dtype = np.double)

	cdef np.ndarray[np.double_t, mode = "c"] background_image_c;
	cdef np.ndarray[np.double_t, mode = "c"] background_color_c;

	if scene.background_image is not None:
		background_image_c =  np.ascontiguousarray(scene.background_image.flatten(), dtype = np.double)
	else:
		background_color_c =  np.ascontiguousarray(scene.background_color.flatten(), dtype = np.double)


	scene_c.height = <int> scene.height
	scene_c.width = <int> scene.width
	scene_c.nb_triangles = nb_triangles
	scene_c.nb_vertices = nb_vertices
	scene_c.backface_culling = scene.backface_culling
	scene_c.clockwise = scene.clockwise	
	scene_c.nb_uv = nb_vertices_uv
	scene_c.faces = <unsigned int*> faces_c.data
	scene_c.faces_uv = <unsigned int*> faces_uv_c.data	
	scene_c.depths = <double*> depths_c.data
	scene_c.uv = <double*> uv_c.data
	scene_c.uv_b = <double*> uv_b_c.data
	scene_c.ij = <double*> ij_c.data
	scene_c.ij_b = <double*> ij_b_c.data
	scene_c.shade = <double*> shade_c.data
	scene_c.shade_b = <double*> shade_b_c.data
	scene_c.colors = <double*> colors_c.data
	scene_c.colors_b = <double*> colors_b_c.data
	scene_c.edgeflags = <bool*> edgeflags_c.data
	scene_c.textured = <bool*> textured_c.data
	scene_c.shaded = <bool*> shaded_c.data
	scene_c.texture = <double*> texture_c.data
	scene_c.texture_b = <double*> texture_b_c.data

	if scene.background_image is not None:	
		scene_c.background_image = <double*> background_image_c.data
		scene_c.background_color = NULL
		if scene_c.background_image  ==  NULL:
			raise BaseException('scene_c.background_image_c is NULL')

	else:
		scene_c.background_image = NULL
		scene_c.background_color = <double*> background_color_c.data
		if scene_c.background_color  ==  NULL:
			raise BaseException('scene_c.background_color is NULL')

	scene_c.texture_height = scene.texture.shape[0]
	scene_c.texture_width = scene.texture.shape[1]
	scene_c.strict_edge = scene.strict_edge
	scene_c.perspective_correct = scene.perspective_correct
	scene_c.integer_pixel_centers = scene.integer_pixel_centers
	

	cdef double* obs_ptr  =  NULL
	cdef double* err_buffer_ptr  =  NULL
	cdef double* err_buffer_b_ptr  =  NULL	
	cdef double* image_ptr  =  <double*> image.data
	
	cdef double* image_b_ptr  =  NULL
	cdef double* z_buffer_ptr  =  <double*> z_buffer.data
	
	if image_ptr  ==  NULL:
		raise BaseException('image_ptr is NULL')		
	if z_buffer_ptr  ==  NULL:
		raise BaseException('z_buffer_ptr is NULL')
	
	if antialiase_error:
		if check_valid:
			assert err_buffer.shape[0]  ==  height 
			assert err_buffer.shape[1]  ==  width 
			assert obs.shape[0]  ==  height 
			assert obs.shape[1]  ==  width 
	
		err_buffer_ptr = <double*>err_buffer.data
		err_buffer_b_ptr = <double*>err_buffer_b.data
		obs_ptr = <double*>obs.data
		
		if err_buffer_ptr  ==  NULL:
			raise BaseException('err_buffer_ptr is NULL')
		if err_buffer_b_ptr  ==  NULL:
			raise BaseException('err_buffer_b_ptr is NULL')
		if obs_ptr  ==  NULL:
			raise BaseException('obs_ptr is NULL')
	else:
		if check_valid:
			assert (not(image_b is None))
			assert image_b.shape[0]  ==  height 
			assert image_b.shape[1]  ==  width 
		image_b_ptr  =  <double*> image_b.data
		if image_b_ptr  ==  NULL:
			raise BaseException('image_b_ptr is NULL')
	
	renderScene_B( scene_c, image_ptr, z_buffer_ptr, image_b_ptr, sigma, antialiase_error ,obs_ptr, err_buffer_ptr, err_buffer_b_ptr)
	scene.uv_b = uv_b_c.reshape(scene.uv_b.shape)
	scene.ij_b = ij_b_c.reshape(scene.ij_b.shape)
	scene.shade_b = shade_b_c.reshape(scene.shade_b.shape)
	scene.colors_b = colors_b_c.reshape(scene.colors_b.shape)
	scene.texture_b = texture_b_c.reshape(scene.texture_b.shape)
	



# type: ignore
"""Module to render deodr scenes using OpenGL through moderngl.

This is used to have a reference implementation using OpenGL shaders that produce identical images to deodr
"""

import moderngl
import numpy as np
from pyrr import Matrix44

from deodr.differentiable_renderer import Camera, Scene3D
from deodr.triangulated_mesh import ColoredTriMesh

from . import shaders as opengl_shaders


def opencv_to_opengl_perspective(
    camera: Camera, znear: float, zfar: float, integer_pixel_centers: bool
) -> np.ndarray:
    # https://blog.noctua-software.com/opencv-opengl-projection-matrix.html
    fx = camera.intrinsic[0, 0]
    fy = camera.intrinsic[1, 1]
    cx = camera.intrinsic[0, 2]
    cy = camera.intrinsic[1, 2]

    if integer_pixel_centers:
        # If integer_pixel_centers is False, then deodr rendering is cnot onsistent with opengl without an 0.5 offset here
        cx2 = cx + 0.5
        cy2 = cy + 0.5
    else:
        # If integer_pixel_centers is False, then deodr rendering is consistent with opengl without need for an offset here
        cx2 = cx
        cy2 = cy

    width = camera.width
    height = camera.height
    np.testing.assert_array_equal(
        [[fx, 0, cx], [0, fy, cy], [0, 0, 1]], camera.intrinsic
    )
    return np.array(
        [
            [2.0 * fx / width, 0, 0, 0],
            [0, -2.0 * fy / height, 0, 0],
            [
                1.0 - 2.0 * cx2 / width,
                1.0 - 2.0 * cy2 / height,
                (zfar + znear) / (znear - zfar),
                -1,
            ],
            [0, 0, 2.0 * zfar * znear / (znear - zfar), 0.0],
        ]
    )


class OffscreenRenderer:
    """Class to perform offscreen rendering of deodr scenes using moderngl."""

    def __init__(self) -> None:
        self.ctx = moderngl.create_standalone_context()

        # Shaders
        self.shader_program = self.ctx.program(
            vertex_shader=opengl_shaders.vertex_shader_source,
            fragment_shader=opengl_shaders.fragment_shader_rgb_source,
        )
        self.fbo = None
        self.texture = None
        self.texture_id: int = 0

    def set_scene(self, deodr_scene: Scene3D) -> None:
        assert deodr_scene.mesh is not None
        assert deodr_scene.light_ambient is not None
        assert deodr_scene.background_color is not None
        self.bg_color = deodr_scene.background_color
        assert deodr_scene.light_directional is not None, "Not supported yet"
        self.set_light(deodr_scene.light_directional, deodr_scene.light_ambient)
        self.set_mesh(deodr_scene.mesh)
        self.integer_pixel_centers = deodr_scene.integer_pixel_centers

    def set_light(self, light_directional: np.ndarray, light_ambient: float) -> None:
        assert light_directional.shape == (3,)
        self.shader_program["light_directional"].value = tuple(
            light_directional
        )  # type :ignore
        self.shader_program["light_ambient"].value = light_ambient

    def set_mesh(self, mesh: ColoredTriMesh) -> None:
        assert mesh.uv is not None
        assert mesh.texture is not None
        self.set_texture(mesh.texture)
        # create triangles soup

        vertices = mesh.vertices[mesh.faces].reshape(-1, 3)
        min_max = np.stack((vertices.min(axis=0), vertices.max(axis=0)))
        self.bounding_box_corners = np.stack(
            np.meshgrid(min_max[:, 0], min_max[:, 1], min_max[:, 2]), axis=-1
        ).reshape(-1, 3)
        normals = mesh.vertex_normals[mesh.faces].reshape(-1, 3)
        uv = mesh.uv[mesh.faces_uv].reshape(-1, 2)
        moderngl_uv = np.column_stack(
            (
                (uv[:, 0] + 0.5) / mesh.texture.shape[1],
                ((uv[:, 1] + 0.5) / mesh.texture.shape[0]),
            )
        )

        vbo_vert = self.ctx.buffer(vertices.astype("f4").tobytes())
        vbo_norm = self.ctx.buffer(normals.astype("f4").tobytes())
        vbo_uv = self.ctx.buffer(moderngl_uv.astype("f4").tobytes())

        self.vao = self.ctx.vertex_array(
            self.shader_program,
            [
                (vbo_vert, "3f", "in_vert"),
                (vbo_norm, "3f", "in_norm"),
                (vbo_uv, "2f", "in_text"),
            ],
        )

    def set_texture(self, texture: np.ndarray) -> None:
        # Texture
        assert not texture.flags["WRITEABLE"]
        assert texture.ndim == 3
        texture_id = id(texture)
        if self.texture is None or self.texture_id != texture_id:
            self.texture_id = id(texture)
            self.texture = self.ctx.texture(
                (texture.shape[1], texture.shape[0]),
                texture.shape[2],
                (texture * 255).astype(np.uint8).tobytes(),
            )
        # texture.build_mipmaps()

    def set_camera(self, camera: Camera) -> None:
        extrinsic = np.row_stack((camera.extrinsic, [0, 0, 0, 1]))

        intrinsic = Matrix44(
            np.diag([1, -1, -1, 1]).dot(
                opencv_to_opengl_perspective(
                    camera,
                    self.znear,
                    self.zfar,
                    integer_pixel_centers=self.integer_pixel_centers,
                )
            )
        )

        #

        self.shader_program["intrinsic"].write(intrinsic.astype("f4").tobytes())
        self.shader_program["extrinsic"].write(extrinsic.T.astype("f4").tobytes())
        if camera.distortion is None:
            k1, k2, p1, p2, k3 = (0.0, 0.0, 0.0, 0.0, 0.0)
        else:
            (
                k1,
                k2,
                p1,
                p2,
                k3,
            ) = camera.distortion
        self.shader_program["k1"].value = k1
        self.shader_program["k2"].value = k2
        self.shader_program["k3"].value = k3
        self.shader_program["p1"].value = p1
        self.shader_program["p2"].value = p2

    def render(self, camera: Camera) -> np.ndarray:
        ctx = self.ctx
        self.zfar = camera.world_to_camera(self.bounding_box_corners)[:, 2].max()
        self.znear = 1e-3 * self.zfar
        # Setting up camera
        self.set_camera(camera)

        # compputing the box around the displaced mesh to get maximum accuracy
        # of the xyz point cloud using unit8 opengl type

        # Framebuffers
        if (
            (self.fbo is None)
            or (self.fbo.height != camera.height)
            or (self.fbo.width != camera.width)
        ):
            self.fbo = ctx.framebuffer(
                ctx.renderbuffer((camera.width, camera.height)),
                ctx.depth_renderbuffer((camera.width, camera.height)),
            )
        assert self.fbo is not None
        # Rendering the RGB image
        self.fbo.use()
        ctx.enable(moderngl.DEPTH_TEST)
        ctx.clear(self.bg_color[0], self.bg_color[1], self.bg_color[2])
        self.texture.use()
        self.vao.render()
        data = self.fbo.read(components=3, alignment=1)
        return np.frombuffer(data, dtype=np.uint8).reshape(
            camera.height, camera.width, 3
        )



"""Test using rgb mesh rendering."""

import hashlib
import os

import imageio.v3 as imageio
import numpy as np

import deodr
from deodr import differentiable_renderer_cython  # type: ignore
from deodr.examples.render_mesh import example_moderngl, example_rgb
from deodr.examples.triangle_soup_fitting import create_example_scene


def test_render_mesh_moderngl() -> None:
    if os.name != "nt":  # did not manage to install mesa on windows github action
        example_moderngl(display=False)


def test_render_mesh_duck(update: bool = False) -> None:
    image = example_rgb(display=False, save_image=False, width=320, height=240)
    image_file = os.path.abspath(os.path.join(deodr.data_path, "test/duck.png"))
    image_uint8 = (image * 255).astype(np.uint8)
    if update:
        imageio.imwrite(image_file, image_uint8)
    image_prev = imageio.imread(image_file)
    assert np.max(np.abs(image_prev.astype(int) - image_uint8.astype(int))) == 0


def test_render_mesh_triangle_soup() -> None:
    np.random.seed(2)
    scene_gt = create_example_scene(clockwise=True)

    assert (
        hashlib.sha256(scene_gt.ij.tobytes()).hexdigest()
        == "56a498bf243bd514c9ab4a3bfd90f8105aa2c168023fa288dc39ad82e2d36a20"
    )
    assert (
        hashlib.sha256(scene_gt.depths.tobytes()).hexdigest()
        == "e25eed6310fef37e401aef594c4c95e1b3cccf962a3646976cf546c58ddfac0a"
    )
    assert (
        hashlib.sha256(scene_gt.uv.tobytes()).hexdigest()
        == "f436623445124ecff7139efa57cce21c2768e23727bac974e236ea33651cc7c9"
    )
    assert (
        hashlib.sha256(scene_gt.shade.tobytes()).hexdigest()
        == "4b796b925c4349245e52a3e6311e99d536dc71e8aa8dc43cbd67cbe35d48892f"
    )
    assert (
        hashlib.sha256(scene_gt.colors.tobytes()).hexdigest()
        == "76dbff728be3eb0860bd27adf493e935dbd81cd7232ec732ba30c4f73ea35c94"
    )

    sigma = 1
    image = np.zeros((scene_gt.height, scene_gt.width, scene_gt.nb_colors))
    z_buffer = np.zeros((scene_gt.height, scene_gt.width))
    differentiable_renderer_cython.renderSceneCpp(scene_gt, sigma, image, z_buffer)

    filename = os.path.join(os.path.dirname(__file__), "data", "triangle_soup.png")
    # imageio.imwrite(filename,image)

    image_lkg = imageio.imread(filename)
    assert np.max(np.abs(image_lkg - image * 255)) <= 1

    if os.name == "nt":  # windows
        assert (
            hashlib.sha256(image.tobytes()).hexdigest()
            == "4de52cc3e902f92ff64324b261ddc45cd6d148ec7e670cf2942532d515af62d8"
        )
        assert (
            hashlib.sha256(z_buffer.tobytes()).hexdigest()
            == "b6f87e03c60bd820efa09d0536495b25d5852f67ecbecd2622f8bf1910d6052a"
        )

    # elif os.name == "posix":  # linux
    #     # google colab  Intel(R) Xeon(R) CPU @ 2.20GHz
    #     assert (
    #         hashlib.sha256(image.tobytes()).hexdigest()
    #         == "ee530428ecac0a11880aa942e92e40515cdebf86a5e6dd7aadc99b8dcaaf11a6"
    #     )
    #     assert (
    #         hashlib.sha256(z_buffer.tobytes()).hexdigest()
    #         == "b6f87e03c60bd820efa09d0536495b25d5852f67ecbecd2622f8bf1910d6052a"
    #     )

    # else:
    #     raise BaseException(f"No results for os.name={os.name}")


if __name__ == "__main__":
    test_render_mesh_duck()
    test_render_mesh_triangle_soup()
    test_render_mesh_moderngl()



"""Module to render deodr scenes using OpenGL through pyrender.

Note that pyrender does not support camera distortion.
"""
from typing import Optional, Tuple

import numpy as np
import pyrender
import trimesh

from deodr.differentiable_renderer import Camera, Scene3D
from deodr.triangulated_mesh import ColoredTriMesh


def arcsinc(x: float) -> float:
    return np.arcsin(x) / x if abs(x) > 1e-6 else 1


def min_rotation(vec1: np.ndarray, vec2: np.ndarray) -> np.ndarray:
    """Find the rotation matrix that aligns vec1 to vec2
    :param vec1: A 3d "source" vector
    :param vec2: A 3d "destination" vector
    :return mat: A transform matrix (3x3) which when applied to vec1,
                 aligns it with vec2.
    """
    assert vec1.shape == (3,)
    assert vec2.shape == (3,)
    a, b = (
        (vec1 / np.linalg.norm(vec1)).reshape(3),
        (vec2 / np.linalg.norm(vec2)).reshape(3),
    )
    v = np.cross(a, b)
    c = np.dot(a, b)
    s = np.linalg.norm(v)
    kmat = np.array([[0, -v[2], v[1]], [v[2], 0, -v[0]], [-v[1], v[0], 0]])
    d = (1 - c) / (s**2) if s > 1e-6 else 0.5
    return np.eye(3) + kmat + kmat.dot(kmat) * d


def deodr_directional_light_to_pyrender(
    deodr_directional_light: np.ndarray,
) -> Tuple[Optional[np.ndarray], Optional[np.ndarray]]:
    assert deodr_directional_light.shape == (3,)
    directional_light_intensity = np.linalg.norm(deodr_directional_light)
    if directional_light_intensity > 0:
        directional_light_direction = (
            deodr_directional_light / directional_light_intensity
        )
        directional_light_rotation = min_rotation(
            np.array([0, 0, -1]), directional_light_direction
        )
        pose = np.zeros((4, 4))
        pose[:3, :3] = directional_light_rotation
        pose[3, 3] = 1
        directional_light = pyrender.light.DirectionalLight(
            intensity=directional_light_intensity
        )
    else:
        directional_light = None
        pose = None

    return directional_light, pose


def deodr_mesh_to_pyrender(deodr_mesh: ColoredTriMesh) -> pyrender.Mesh:
    assert deodr_mesh.uv is not None
    assert deodr_mesh.texture is not None

    # trimesh and pyrender do to handle faces indices for texture
    # that are different from face indices for the 3d vertices
    # we need to duplicate vertices
    faces, mask_v, mask_vt = trimesh.visual.texture.unmerge_faces(
        deodr_mesh.faces, deodr_mesh.faces_uv
    )
    vertices = deodr_mesh.vertices[mask_v]
    deodr_mesh.compute_vertex_normals()
    vertex_normals = deodr_mesh.vertex_normals[mask_v]
    uv = deodr_mesh.uv[mask_vt]
    pyrender_uv = np.column_stack(
        (
            (uv[:, 0]) / deodr_mesh.texture.shape[0],
            (1 - uv[:, 1] / deodr_mesh.texture.shape[1]),
        )
    )

    material = None
    poses = None
    color_0 = None
    if material is None:
        base_color_texture = pyrender.texture.Texture(
            source=deodr_mesh.texture, source_channels="RGB"
        )
        material = pyrender.MetallicRoughnessMaterial(
            alphaMode="BLEND",
            baseColorFactor=[1, 1, 1, 1.0],
            metallicFactor=0,
            roughnessFactor=1,
            baseColorTexture=base_color_texture,
        )
        material.wireframe = False

    primitive = pyrender.Primitive(
        positions=vertices,
        normals=vertex_normals,
        texcoord_0=pyrender_uv,
        color_0=color_0,
        indices=faces,
        material=material,
        mode=pyrender.constants.GLTF.TRIANGLES,
        poses=poses,
    )

    return pyrender.Mesh(primitives=[primitive])


def render(deodr_scene: Scene3D, camera: Camera) -> Tuple[np.ndarray, np.ndarray]:
    assert (
        deodr_scene.mesh is not None
    ), "You need a mesh in the scene you want to render"
    """Render a deodr scene using pyrender"""
    if deodr_scene.background_image is not None:
        bg_color = deodr_scene.background_image[0, 0]
        if not (np.all(deodr_scene.background_image == bg_color[None, None, :])):
            raise (
                BaseException(
                    "pyrender does not support background image, please provide a"
                    " background image that correspond to a uniform color"
                )
            )
    else:
        bg_color = deodr_scene.background_color
    pyrender_scene = pyrender.Scene(
        ambient_light=deodr_scene.light_ambient * np.ones((3)), bg_color=bg_color
    )

    if camera.distortion is not None:
        raise (BaseException("cameras with distortion not handled yet with pyrender"))

    # cam = pyrender.PerspectiveCamera(yfov=np.pi / 3.0, aspectRatio=1.414)
    cam = pyrender.IntrinsicsCamera(
        fx=camera.intrinsic[0, 0],
        fy=camera.intrinsic[1, 1],
        cx=camera.intrinsic[0, 2],
        cy=camera.intrinsic[1, 2],
    )

    # convert to pyrender directional light parameterization
    assert deodr_scene.light_directional is not None
    directional_light, directional_light_pose = deodr_directional_light_to_pyrender(
        deodr_scene.light_directional
    )
    if directional_light is not None:
        pyrender_scene.add(directional_light, pose=directional_light_pose)

    # convert the mesh
    pyrender_mesh = deodr_mesh_to_pyrender(deodr_scene.mesh)
    pyrender_scene.add(pyrender_mesh, pose=np.eye(4))

    # convert the camera
    m = camera.camera_to_world_mtx_4x4()
    pose_camera = m.dot(np.diag([1, -1, -1, 1]))

    pyrender_scene.add(cam, pose=pose_camera)

    # render
    width = camera.width
    height = camera.height
    r = pyrender.offscreen.OffscreenRenderer(width, height, point_size=1.0)
    image_pyrender, depth = r.render(pyrender_scene)
    return image_pyrender, depth



"""Example of fitting a hand mesh to a depth image."""

import datetime
import glob
import json
import os
import time
from typing import List

import cv2
from imageio.v3 import imwrite
import matplotlib.pyplot as plt
import numpy as np

import deodr
from deodr import ColoredTriMesh
from deodr.mesh_fitter import MeshDepthFitter
from deodr.pytorch import MeshDepthFitter as PytorchMeshDepthFitter
from deodr.tensorflow import MeshDepthFitter as TensorFlowMeshDepthFitter


def run(
    dl_library: str = "none",
    plot_curves: bool = False,
    save_images: bool = False,
    display: bool = True,
    max_iter: int = 300,
    n_subdivision: int = 0,
) -> List[float]:
    file_folder = os.path.dirname(__file__)

    depth_image = np.fliplr(
        np.fromfile(os.path.join(deodr.data_path, "depth.bin"), dtype=np.float32)
        .reshape(240, 320)
        .astype(np.float64)
    )
    depth_image = depth_image[20:-20, 60:-60]
    max_depth = 450
    depth_image[depth_image == 0] = max_depth
    depth_image = depth_image / max_depth

    obj_file = os.path.join(deodr.data_path, "hand.obj")
    faces, vertices = deodr.read_obj(obj_file)
    mesh = ColoredTriMesh(faces.copy(), vertices=vertices, nb_colors=0).subdivise(
        n_subdivision
    )

    euler_init = np.array([0.1, 0.1, 0.1])
    translation_init = np.zeros(3)

    MeshDepthFittersSelector = {
        "none": MeshDepthFitter,
        "pytorch": PytorchMeshDepthFitter,
        "tensorflow": TensorFlowMeshDepthFitter,
    }

    hand_fitter: MeshDepthFitter = MeshDepthFittersSelector[dl_library](  # type: ignore
        mesh.vertices, mesh.faces, euler_init, translation_init, cregu=1000
    )
    distortion = np.array([1, 0, 0, 0, 0])
    hand_fitter.set_image(depth_image, focal=241, distortion=distortion)
    hand_fitter.set_max_depth(1)
    hand_fitter.set_depth_scale(110 / max_depth)
    energies: List[float] = []
    durations: List[float] = []
    start = time.time()

    iter_folder = os.path.join(file_folder, "./iterations/depth")
    if not os.path.exists(iter_folder):
        os.makedirs(iter_folder)

    for niter in range(max_iter):
        energy, synthetic_depth, diff_image = hand_fitter.step()
        energies.append(energy)
        durations.append(time.time() - start)
        if save_images or display:
            combined_image = np.column_stack(
                (depth_image, synthetic_depth, 3 * diff_image)
            )
        if display:
            cv2.imshow("animation", cv2.resize(combined_image, None, fx=2, fy=2))
        if save_images:
            imwrite(
                os.path.join(iter_folder, f"depth_hand_iter_{niter}.png"),
                combined_image,
            )
        cv2.waitKey(1)

    with open(
        os.path.join(
            iter_folder,
            f'depth_image_fitting_result_{str(datetime.datetime.now()).replace(":", "_")}.json',
        ),
        "w",
    ) as f:
        json.dump(
            {
                "label": f"{dl_library} {datetime.datetime.now()}",
                "durations": durations,
                "energies": energies,
            },
            f,
            indent=4,
        )

    if plot_curves:
        plt.figure()
        for file in glob.glob(
            os.path.join(iter_folder, "depth_image_fitting_result_*.json")
        ):
            with open(file, "r") as fp:
                json_data = json.load(fp)
                plt.plot(
                    json_data["durations"],
                    json_data["energies"],
                    label=json_data["label"],
                )
        plt.legend()
        plt.figure()
        for file in glob.glob(
            os.path.join(iter_folder, "depth_image_fitting_result_*.json")
        ):
            with open(file, "r") as fp:
                json_data = json.load(fp)
                plt.plot(json_data["energies"], label=json_data["label"])
        plt.legend()
        plt.show()

    return energies


def main() -> None:
    display = True
    n_subdivision = 0
    run(
        dl_library="none",
        plot_curves=False,
        save_images=False,
        display=display,
        n_subdivision=n_subdivision,
    )

    run(
        dl_library="pytorch",
        plot_curves=False,
        save_images=False,
        display=display,
        n_subdivision=n_subdivision,
    )

    run(
        dl_library="tensorflow",
        plot_curves=True,
        save_images=False,
        display=display,
        n_subdivision=n_subdivision,
    )


if __name__ == "__main__":
    main()



"""Pytorch interface for deodr."""

__all__ = [
    "ColoredTriMeshPytorch",
    "Scene3DPytorch",
    "CameraPytorch",
    "LaplacianRigidEnergyPytorch",
    "MeshRGBFitterWithPose",
    "MeshDepthFitter",
]


from .differentiable_renderer_pytorch import CameraPytorch  # type: ignore
from .differentiable_renderer_pytorch import Scene3DPytorch  # type: ignore
from .laplacian_rigid_energy_pytorch import LaplacianRigidEnergyPytorch  # type: ignore
from .mesh_fitter_pytorch import MeshDepthFitter  # type: ignore
from .mesh_fitter_pytorch import MeshRGBFitterWithPose  # type: ignore
from .triangulated_mesh_pytorch import ColoredTriMeshPytorch  # type: ignore



"""Example with fitting a colored hand mesh model to an image."""
import datetime
import glob
import json
import os
import time
from typing import List

import cv2
from imageio.v3 import imread, imwrite
import matplotlib.pyplot as plt
import numpy as np
from typing_extensions import Literal

import deodr
from deodr import ColoredTriMesh, read_obj
from deodr.mesh_fitter import MeshRGBFitterWithPose
from deodr.meshlab_io import export_meshlab
from deodr.pytorch import MeshRGBFitterWithPose as PyTorchMeshRGBFitterWithPose
from deodr.tensorflow import (
    MeshRGBFitterWithPose as TensorflowTorchMeshRGBFitterWithPose,
)

DlLibraryType = Literal["pytorch", "tensorflow", "none"]


def run(
    dl_library: DlLibraryType = "pytorch",
    plot_curves: bool = True,
    save_images: bool = True,
    display: bool = True,
    max_iter: int = 100,
    n_subdivision: int = 0,
) -> List[float]:
    MeshFittersSelector = {
        "none": MeshRGBFitterWithPose,
        "pytorch": PyTorchMeshRGBFitterWithPose,
        "tensorflow": TensorflowTorchMeshRGBFitterWithPose,
    }

    hand_image = (
        imread(os.path.join(deodr.data_path, "hand.png")).astype(np.double) / 255
    )

    obj_file = os.path.join(deodr.data_path, "hand.obj")
    faces, vertices = read_obj(obj_file)

    mesh = ColoredTriMesh(faces.copy(), vertices=vertices, nb_colors=3).subdivise(
        n_subdivision
    )

    default_color = np.array([0.4, 0.3, 0.25])
    default_light_directional = -np.array([0.1, 0.5, 0.4])
    default_light_ambient = 0.6
    euler_init = np.array([0, 0, 0])
    translation_init = np.mean(mesh.vertices, axis=0)
    # centering vertices
    mesh.set_vertices(mesh.vertices - translation_init[None, :])

    hand_fitter: MeshRGBFitterWithPose = MeshFittersSelector[dl_library](  # type: ignore
        mesh.vertices,
        mesh.faces,
        default_color=default_color,
        default_light_directional=default_light_directional,
        default_light_ambient=default_light_ambient,
        update_lights=True,
        update_color=True,
        euler_init=euler_init,
        translation_init=translation_init,
        cregu=1000,
    )

    hand_fitter.reset()

    background_color = np.median(
        np.row_stack(
            (
                hand_image[:10, :10, :].reshape(-1, 3),
                hand_image[-10:, :10, :].reshape(-1, 3),
                hand_image[-10:, -10:, :].reshape(-1, 3),
                hand_image[:10, -10:, :].reshape(-1, 3),
            )
        ),
        axis=0,
    )

    background_color = np.array([0.5, 0.6, 0.7])
    distortion = np.array([-1, 0, 0, 0, 0])
    hand_fitter.set_image(hand_image, distortion=distortion)
    hand_fitter.set_background_color(background_color)
    energies = []
    durations = []
    start = time.time()

    iterfolder = "./iterations/rgb"
    if not os.path.exists(iterfolder):
        os.makedirs(iterfolder)

    for niter in range(max_iter):
        energy, image, diff_image = hand_fitter.step()
        energies.append(energy)
        durations.append(time.time() - start)
        if display or save_images:
            combined_image = np.column_stack(
                (hand_image, image, np.tile(diff_image[:, :, None], (1, 1, 3)))
            )
        if display:
            cv2.imshow(
                "animation",
                cv2.resize(combined_image[:, :, ::-1], None, fx=2, fy=2),
            )
        if save_images:
            imwrite(os.path.join(iterfolder, f"hand_iter_{niter}.png"), combined_image)
        cv2.waitKey(1)

    export_meshlab(
        "iterations/rgb_fitted_meshlab.mlp",
        hand_fitter.mesh,
        [hand_fitter.camera],
        [hand_image],
    )
    # save convergence curve
    with open(
        os.path.join(
            iterfolder,
            f'rgb_image_fitting_result_{str(datetime.datetime.now()).replace(":", "_")}.json',
        ),
        "w",
    ) as f:
        json.dump(
            {
                "label": f"{dl_library} {datetime.datetime.now()}",
                "durations": durations,
                "energies": energies,
            },
            f,
            indent=4,
        )

    # compare with previous runs
    if plot_curves:
        plt.figure()
        for file in glob.glob(
            os.path.join(iterfolder, "rgb_image_fitting_result_*.json")
        ):
            with open(file, "r") as fp:
                json_data = json.load(fp)
                plt.plot(
                    json_data["durations"],
                    json_data["energies"],
                    label=json_data["label"],
                )
                plt.xlabel("duration in seconds")
                plt.ylabel("energies")
        plt.legend()
        plt.figure()
        for file in glob.glob(
            os.path.join(iterfolder, "rgb_image_fitting_result_*.json")
        ):
            with open(file, "r") as fp:
                json_data = json.load(fp)
                plt.plot(json_data["energies"], label=json_data["label"])
                plt.xlabel("interations")
                plt.ylabel("energies")
        plt.legend()
        plt.show()
    return energies


def main() -> None:
    display = True
    save_images = False
    n_subdivision = 1

    run(
        dl_library="pytorch",
        plot_curves=False,
        display=display,
        save_images=save_images,
        n_subdivision=n_subdivision,
    )

    run(
        dl_library="none",
        plot_curves=False,
        display=display,
        save_images=save_images,
        n_subdivision=n_subdivision,
    )
    run(
        dl_library="tensorflow",
        plot_curves=True,
        display=display,
        save_images=save_images,
        n_subdivision=n_subdivision,
    )


if __name__ == "__main__":
    main()



# type: ignore
"""Pytorch interface to deodr."""

from typing import Any, List, Optional, Tuple, Union

import numpy as np
import torch

from .. import differentiable_renderer_cython  # type: ignore
from ..differentiable_renderer import Camera, Scene3D


class CameraPytorch(Camera):
    """Pytorch implementation of the camera class."""

    def __init__(
        self,
        extrinsic: np.ndarray,
        intrinsic: np.ndarray,
        height: int,
        width: int,
        distortion: Optional[np.ndarray] = None,
    ):
        super().__init__(
            extrinsic, intrinsic, height, width, distortion=distortion, checks=False
        )

    def world_to_camera(self, points_3d: torch.Tensor) -> torch.Tensor:
        assert isinstance(points_3d, torch.Tensor)
        return torch.cat(
            (points_3d, torch.ones((points_3d.shape[0], 1), dtype=torch.double)), dim=1
        ).mm(torch.tensor(self.extrinsic.T))

    def left_mul_intrinsic(self, projected: torch.Tensor) -> torch.Tensor:
        return torch.cat(
            (projected, torch.ones((projected.shape[0], 1), dtype=torch.double)), dim=1
        ).mm(torch.tensor(self.intrinsic[:2, :].T))

    def column_stack(
        self, values: Union[List[torch.Tensor], Tuple[torch.Tensor, ...]]
    ) -> torch.Tensor:
        return torch.stack(values, dim=1)


class TorchDifferentiableRenderer2DFunc(torch.autograd.Function):
    """Pytorch implementation of the 2D rendering function."""

    @staticmethod
    def forward(  # type: ignore
        ctx: Any, ij: torch.Tensor, colors: torch.Tensor, scene: "Scene3DPytorch"
    ) -> torch.Tensor:
        nb_color_channels = colors.shape[1]
        image = np.empty(
            (scene.scene_2d.height, scene.scene_2d.width, nb_color_channels)
        )
        z_buffer = np.empty((scene.scene_2d.height, scene.scene_2d.width))
        ctx.scene = scene
        scene.scene_2d.ij = (
            ij.detach().numpy()
        )  # should automatically detached according to
        # https://pytorch.org/docs/master/notes/extending.html
        scene.colors = colors.detach().numpy()
        differentiable_renderer_cython.renderSceneCpp(
            scene.scene_2d, 1, image, z_buffer
        )
        ctx.save_for_backward(ij, colors)
        ctx.image = image.copy()
        # making a copy to keep the antializaed image for visualization ,
        # could be optional
        ctx.z_buffer = z_buffer
        return torch.as_tensor(image)

    @staticmethod
    def backward(ctx: Any, *grad_outputs: Any) -> Any:
        assert len(grad_outputs) == 1
        image_b = grad_outputs[0]
        scene = ctx.scene
        scene.scene_2d.uv_b = np.zeros(scene.scene_2d.uv.shape)
        scene.scene_2d.ij_b = np.zeros(scene.scene_2d.ij.shape)
        scene.scene_2d.shade_b = np.zeros(scene.scene_2d.shade.shape)
        scene.scene_2d.colors_b = np.zeros(scene.scene_2d.colors.shape)
        scene.scene_2d.texture_b = np.zeros(scene.scene_2d.texture.shape)
        differentiable_renderer_cython.renderSceneBCpp(
            scene.scene_2d, 1, ctx.image, ctx.z_buffer, image_b.numpy()
        )
        return (
            torch.as_tensor(scene.scene_2d.ij_b),
            torch.as_tensor(scene.scene_2d.colors_b),
            None,
        )


TorchDifferentiableRender2D = TorchDifferentiableRenderer2DFunc.apply


class Scene3DPytorch(Scene3D):
    """Pytorch implementation of deodr 3D scenes."""

    def __init__(self) -> None:
        super().__init__()

    def set_light(
        self, light_directional: torch.Tensor, light_ambient: torch.Tensor
    ) -> None:
        if not (isinstance(light_directional, torch.Tensor)):
            light_directional = torch.tensor(light_directional)
        self.light_directional = light_directional
        self.light_ambient_pytorch = light_ambient

    def _compute_vertices_colors_with_illumination(self) -> torch.Tensor:
        assert self.mesh is not None
        vertices_luminosity = (
            torch.relu(
                -torch.sum(self.mesh.vertex_normals * self.light_directional, dim=1)
            )
            + self.light_ambient_pytorch
        )
        return self.mesh.vertices_colors * vertices_luminosity[:, None]

    def _render_2d(self) -> Tuple[torch.Tensor, torch.Tensor]:
        self.depths = self.scene_2d.depths.detach()
        return (
            TorchDifferentiableRender2D(self.scene_2d.ij, self.scene_2d.colors, self),
            self.depths,
        )



# type: ignore
"""Pytorch implementation of a triangulated mesh."""

from typing import Callable, Optional

import numpy as np
import torch
from torch.sparse import DoubleTensor  # type: ignore

from ..triangulated_mesh import ColoredTriMesh, TriMeshAdjacencies


def print_grad(name: str) -> Callable[[torch.Tensor], None]:
    def hook(grad: torch.Tensor) -> None:
        print(f"grad {name} = {grad}")

    return hook


class TriMeshAdjacenciesPytorch(TriMeshAdjacencies):
    """Class that stores adjacency matrices and methods that use this adjacencies using pytorch sparse matrices.
    Unlike the TriMesh class there are no vertices stored in this class.
    """

    def __init__(self, faces: np.ndarray, clockwise: bool = False):
        super().__init__(faces, clockwise)
        self.faces_torch = torch.LongTensor(faces)
        i = self.faces_torch.flatten()
        j = torch.LongTensor(
            np.tile(np.arange(self.nb_faces)[:, None], [1, 3]).flatten()
        )
        self._vertices_faces_torch = DoubleTensor(
            torch.stack((i, j)),
            torch.ones((self.nb_faces, 3), dtype=torch.float64).flatten(),
            torch.Size((self.nb_vertices, self.nb_faces)),
        )

    def compute_face_normals(self, vertices: torch.Tensor) -> torch.Tensor:
        triangles = vertices[self.faces_torch, :]
        u = triangles[::, 1] - triangles[::, 0]
        v = triangles[::, 2] - triangles[::, 0]
        n = -torch.cross(u, v) if self.clockwise else torch.cross(u, v)
        l2 = (n**2).sum(dim=1)
        norm = l2.sqrt()
        return n / norm[:, None]

    def compute_vertex_normals(self, face_normals: torch.Tensor) -> torch.Tensor:
        n = self._vertices_faces_torch.mm(face_normals)
        l2 = (n**2).sum(dim=1)
        norm = l2.sqrt()
        return n / norm[:, None]

    def edge_on_silhouette(self, vertices_2d: torch.Tensor) -> np.ndarray:
        return super().edge_on_silhouette(vertices_2d.detach().numpy())


class ColoredTriMeshPytorch(ColoredTriMesh):
    """Pytorch implementation of colored a triangulated mesh."""

    def __init__(
        self,
        faces: np.ndarray,
        vertices: np.ndarray,
        clockwise: bool = False,
        faces_uv: Optional[np.ndarray] = None,
        uv: Optional[np.ndarray] = None,
        texture: Optional[np.ndarray] = None,
        colors: Optional[np.ndarray] = None,
    ):
        super(ColoredTriMeshPytorch, self).__init__(
            faces,
            vertices=vertices,
            clockwise=clockwise,
            faces_uv=faces_uv,
            uv=uv,
            texture=texture,
            colors=colors,
        )

    def compute_adjacencies(self) -> None:
        self._adjacencies = TriMeshAdjacenciesPytorch(self.faces)

    def set_vertices_colors(self, colors: torch.Tensor) -> None:
        self.vertices_colors = colors



"""Modules containing classes to fit 3D meshes to images using differentiable rendering."""

import copy
from typing import Any, Dict, List, Optional, Tuple

import numpy as np
import scipy.sparse.linalg
import scipy.spatial.transform.rotation

from . import Camera, ColoredTriMesh, LaplacianRigidEnergy, Scene3D
from .tools import (
    check_jacobian_finite_differences,
    normalize,
    normalize_backward,
    qrot,
    qrot_backward,
)


class MeshDepthFitter:
    """Class to fit a deformable mesh to a depth image."""

    def __init__(
        self,
        vertices: np.ndarray,
        faces: np.ndarray,
        euler_init: np.ndarray,
        translation_init: np.ndarray,
        cregu: float = 2000,
        inertia: float = 0.96,
        damping: float = 0.05,
    ):
        self.cregu = cregu
        self.inertia = inertia
        self.damping = damping
        self.step_factor_vertices = 0.0005
        self.step_max_vertices = 1
        self.step_factor_quaternion = 0.00006
        self.step_max_quaternion = 0.1
        self.step_factor_translation = 0.00005
        self.step_max_translation = 0.1

        self.mesh = ColoredTriMesh(
            faces, vertices=vertices, colors=np.zeros((vertices.shape[0], 0))
        )  # we do a copy to avoid negative stride not support by pytorch
        object_center = vertices.mean(axis=0)
        object_radius = np.max(np.std(vertices, axis=0))
        self.camera_center = object_center + np.array([-0.5, 0, 5]) * object_radius

        self.scene = Scene3D()
        self.scene.set_mesh(self.mesh)
        self.rigid_energy = LaplacianRigidEnergy(self.mesh, vertices, cregu)
        self.vertices_init = copy.copy(vertices)
        self.Hfactorized = None
        self.Hpreconditioner = None
        self.set_mesh_transform_init(euler=euler_init, translation=translation_init)

        self.reset()

    def set_mesh_transform_init(
        self, euler: np.ndarray, translation: np.ndarray
    ) -> None:
        self.transform_quaternion_init = scipy.spatial.transform.Rotation.from_euler(
            "zyx", euler
        ).as_quat()
        self.transform_translation_init = translation

    def reset(self) -> None:
        self.vertices = copy.copy(self.vertices_init)
        self.speed_vertices = np.zeros(self.vertices_init.shape)
        self.transform_quaternion = copy.copy(self.transform_quaternion_init)
        self.transform_translation = copy.copy(self.transform_translation_init)
        self.speed_translation = np.zeros(3)
        self.speed_quaternion = np.zeros(4)

    def set_max_depth(self, max_depth: float) -> None:
        self.max_depth = max_depth
        self.scene.set_background_color(np.array([max_depth], dtype=np.float64))

    def set_depth_scale(self, depth_scale: float) -> None:
        self.depthScale = depth_scale

    def set_image(
        self,
        mesh_image: np.ndarray,
        focal: Optional[float] = None,
        distortion: Optional[np.ndarray] = None,
    ) -> None:
        self.width = mesh_image.shape[1]
        self.height = mesh_image.shape[0]
        assert mesh_image.ndim == 2
        self.mesh_image = mesh_image
        if focal is None:
            focal = 2 * self.width
        rot = np.array([[1, 0, 0], [0, -1, 0], [0, 0, -1]])
        trans = -rot.T.dot(self.camera_center)
        intrinsic = np.array(
            [[focal, 0, self.width / 2], [0, focal, self.height / 2], [0, 0, 1]]
        )
        extrinsic = np.column_stack((rot, trans))
        self.camera = Camera(
            extrinsic=extrinsic,
            intrinsic=intrinsic,
            distortion=distortion,
            height=self.height,
            width=self.width,
        )
        self.iter = 0

    def render(self) -> np.ndarray:
        q_normalized = normalize(
            self.transform_quaternion
        )  # that will lead to a gradient that is in the tangent space
        vertices_transformed = (
            qrot(q_normalized, self.vertices) + self.transform_translation
        )
        self.mesh.set_vertices(vertices_transformed)
        self.depth_not_clipped = self.scene.render_depth(
            self.camera,
            depth_scale=self.depthScale,
        )
        return np.clip(self.depth_not_clipped, 0, self.max_depth)

    def render_backward(self, depth_b: np.ndarray) -> None:
        self.scene.clear_gradients()
        depth_b[self.depth_not_clipped < 0] = 0
        depth_b[self.depth_not_clipped > self.max_depth] = 0
        self.scene.render_depth_backward(depth_b)
        assert self.scene.mesh is not None
        assert self.scene.mesh._vertices_b is not None
        vertices_transformed_b = self.scene.mesh._vertices_b
        self.transform_translation_b = np.sum(vertices_transformed_b, axis=0)
        q_normalized = normalize(self.transform_quaternion)
        q_normalized_b, self._vertices_b = qrot_backward(
            q_normalized, self.vertices, vertices_transformed_b
        )
        self.transform_quaternion_b = normalize_backward(
            self.transform_quaternion, q_normalized_b
        )  # that will lead to a gradient that is in the tangeant space

    def step(self) -> Tuple[float, np.ndarray, np.ndarray]:
        self.vertices = self.vertices - np.mean(self.vertices, axis=0)[None, :]
        depth = self.render()

        diff_image = np.sum((depth - self.mesh_image[:, :, None]) ** 2, axis=2)
        energy_data = np.sum(diff_image)
        depth_b = 2 * (depth - self.mesh_image[:, :, None])
        self.render_backward(depth_b)

        self._vertices_b = self._vertices_b - np.mean(self._vertices_b, axis=0)[None, :]
        grad_data = self._vertices_b
        # update v

        (
            energy_rigid,
            grad_rigidity,
            _,
        ) = self.rigid_energy.evaluate(self.vertices)
        energy = energy_data + energy_rigid
        print("Energy=%f : EData=%f E_rigid=%f" % (energy, energy_data, energy_rigid))

        # update v
        grad = grad_data + grad_rigidity

        def mult_and_clamp(x: np.ndarray, a: float, t: float) -> np.ndarray:
            return np.minimum(np.maximum(x * a, -t), t)

        inertia = self.inertia
        # update vertices
        step_vertices = mult_and_clamp(
            -grad, self.step_factor_vertices, self.step_max_vertices
        )
        self.speed_vertices = (1 - self.damping) * (
            self.speed_vertices * self.inertia + (1 - self.inertia) * step_vertices
        )
        self.vertices = self.vertices + self.speed_vertices
        # update rotation
        step_quaternion = mult_and_clamp(
            -self.transform_quaternion_b,
            self.step_factor_quaternion,
            self.step_max_quaternion,
        )
        self.speed_quaternion = (1 - self.damping) * (
            self.speed_quaternion * inertia + (1 - inertia) * step_quaternion
        )
        self.transform_quaternion = self.transform_quaternion + self.speed_quaternion
        self.transform_quaternion = self.transform_quaternion / np.linalg.norm(
            self.transform_quaternion
        )
        # update translation
        step_translation = mult_and_clamp(
            -self.transform_translation_b,
            self.step_factor_translation,
            self.step_max_translation,
        )
        self.speed_translation = (1 - self.damping) * (
            self.speed_translation * inertia + (1 - inertia) * step_translation
        )
        self.transform_translation = self.transform_translation + self.speed_translation

        self.iter += 1
        return energy, depth[:, :, 0], diff_image


class MeshRGBFitterWithPose:
    """Class to fit a deformable mesh to a color image."""

    def __init__(
        self,
        vertices: np.ndarray,
        faces: np.ndarray,
        euler_init: np.ndarray,
        translation_init: np.ndarray,
        default_color: np.ndarray,
        default_light_directional: np.ndarray,
        default_light_ambient: float,
        cregu: float = 2000,
        inertia: float = 0.96,
        damping: float = 0.05,
        update_lights: bool = True,
        update_color: bool = True,
    ):
        self.cregu = cregu

        self.inertia = inertia
        self.damping = damping
        self.step_factor_vertices = 0.0005
        self.step_max_vertices = 0.5
        self.step_factor_quaternion = 0.00006
        self.step_max_quaternion = 0.05
        self.step_factor_translation = 0.00005
        self.step_max_translation = 0.1

        self.default_color = default_color
        self.default_light_directional = default_light_directional
        self.default_light_ambient = default_light_ambient
        self.update_lights = update_lights
        self.update_color = update_color
        self.mesh = ColoredTriMesh(faces.copy(), vertices=vertices, nb_colors=3)
        object_center = vertices.mean(axis=0) + translation_init
        object_radius = np.max(np.std(vertices, axis=0))
        self.camera_center = object_center + np.array([0, 0, 9]) * object_radius

        self.scene = Scene3D()
        self.scene.set_mesh(self.mesh)
        self.rigid_energy = LaplacianRigidEnergy(self.mesh, vertices, cregu)
        self.vertices_init = copy.copy(vertices)
        self.Hfactorized = None
        self.Hpreconditioner = None
        self.set_mesh_transform_init(euler=euler_init, translation=translation_init)
        self.reset()

    def set_background_color(self, background_color: np.ndarray) -> None:
        self.scene.set_background_color(background_color)

    def set_mesh_transform_init(
        self, euler: np.ndarray, translation: np.ndarray
    ) -> None:
        self.transform_quaternion_init = scipy.spatial.transform.Rotation.from_euler(
            "zyx", euler
        ).as_quat()
        self.transform_translation_init = translation

    def reset(self) -> None:
        self.vertices = copy.copy(self.vertices_init)
        self.speed_vertices = np.zeros(self.vertices.shape)
        self.transform_quaternion = copy.copy(self.transform_quaternion_init)
        self.transform_translation = copy.copy(self.transform_translation_init)
        self.speed_translation = np.zeros(3)
        self.speed_quaternion = np.zeros(4)

        self.mesh_color = copy.copy(self.default_color)
        self.light_directional = copy.copy(self.default_light_directional)
        self.light_ambient = copy.copy(self.default_light_ambient)

        self.speed_light_directional = np.zeros(self.light_directional.shape)
        self.speed_light_ambient = 0.0
        self.speed_mesh_color = np.zeros(self.mesh_color.shape)

    def set_image(
        self,
        mesh_image: np.ndarray,
        focal: Optional[float] = None,
        distortion: Optional[np.ndarray] = None,
    ) -> None:
        self.width = mesh_image.shape[1]
        self.height = mesh_image.shape[0]
        assert mesh_image.ndim == 3
        self.mesh_image = mesh_image
        if focal is None:
            focal = 2 * self.width

        rot = np.array([[1, 0, 0], [0, -1, 0], [0, 0, -1]])
        trans = -rot.T.dot(self.camera_center)
        intrinsic = np.array(
            [[focal, 0, self.width / 2], [0, focal, self.height / 2], [0, 0, 1]]
        )
        extrinsic = np.column_stack((rot, trans))
        self.camera = Camera(
            extrinsic=extrinsic,
            intrinsic=intrinsic,
            distortion=distortion,
            width=self.width,
            height=self.height,
        )
        self.iter = 0

    def render(self) -> np.ndarray:
        q_normalized = normalize(
            self.transform_quaternion
        )  # that will lead to a gradient that is in the tangent space
        vertices_transformed = (
            qrot(q_normalized, self.vertices) + self.transform_translation
        )
        self.mesh.set_vertices(vertices_transformed)
        self.scene.set_light(
            light_directional=self.light_directional, light_ambient=self.light_ambient
        )
        self.mesh.set_vertices_colors(
            np.tile(self.mesh_color, (self.mesh.nb_vertices, 1))
        )
        return self.scene.render(self.camera)

    def render_backward(self, image_b: np.ndarray) -> None:
        assert self.scene.mesh is not None
        self.scene.clear_gradients()
        self.scene.render_backward(image_b)
        assert self.mesh.vertices_colors_b is not None  # helping mypy
        self.mesh_color_b = np.sum(self.mesh.vertices_colors_b, axis=0)
        self.light_directional_b = self.scene.light_directional_b
        self.light_ambient_b = self.scene.light_ambient_b
        vertices_transformed_b = self.scene.mesh._vertices_b
        self.transform_translation_b = np.sum(vertices_transformed_b, axis=0)
        q_normalized = normalize(self.transform_quaternion)
        q_normalized_b, self._vertices_b = qrot_backward(
            q_normalized, self.vertices, vertices_transformed_b
        )
        self.transform_quaternion_b = normalize_backward(
            self.transform_quaternion, q_normalized_b
        )  # that will lead to a gradient that is in the tangent space

    def step(self) -> Tuple[float, np.ndarray, np.ndarray]:
        self.vertices = self.vertices - np.mean(self.vertices, axis=0)[None, :]

        image = self.render()

        diff_image = np.sum((image - self.mesh_image) ** 2, axis=2)
        image_b = 2 * (image - self.mesh_image)
        energy_data = np.sum(diff_image)

        (
            energy_rigid,
            grad_rigidity,
            _,
        ) = self.rigid_energy.evaluate(self.vertices)
        energy = energy_data + energy_rigid
        print("Energy=%f : EData=%f E_rigid=%f" % (energy, energy_data, energy_rigid))

        self.render_backward(image_b)

        self._vertices_b = self._vertices_b - np.mean(self._vertices_b, axis=0)[None, :]
        # update v
        grad = self._vertices_b + grad_rigidity

        def mult_and_clamp(x: np.ndarray, a: float, t: float) -> np.ndarray:
            return np.minimum(np.maximum(x * a, -t), t)

        inertia = self.inertia

        # update vertices
        step_vertices = mult_and_clamp(
            -grad, self.step_factor_vertices, self.step_max_vertices
        )
        self.speed_vertices = (1 - self.damping) * (
            self.speed_vertices * inertia + (1 - inertia) * step_vertices
        )
        self.vertices = self.vertices + self.speed_vertices
        # update rotation
        step_quaternion = mult_and_clamp(
            -self.transform_quaternion_b,
            self.step_factor_quaternion,
            self.step_max_quaternion,
        )
        self.speed_quaternion = (1 - self.damping) * (
            self.speed_quaternion * inertia + (1 - inertia) * step_quaternion
        )
        self.transform_quaternion = self.transform_quaternion + self.speed_quaternion
        self.transform_quaternion = self.transform_quaternion / np.linalg.norm(
            self.transform_quaternion
        )
        # update translation
        step_translation = mult_and_clamp(
            -self.transform_translation_b,
            self.step_factor_translation,
            self.step_max_translation,
        )
        self.speed_translation = (1 - self.damping) * (
            self.speed_translation * inertia + (1 - inertia) * step_translation
        )
        self.transform_translation = self.transform_translation + self.speed_translation
        # update directional light
        step = -self.light_directional_b * 0.0001
        self.speed_light_directional = (1 - self.damping) * (
            self.speed_light_directional * inertia + (1 - inertia) * step
        )
        self.light_directional = self.light_directional + self.speed_light_directional
        # update ambient light
        step = -self.light_ambient_b * 0.0001
        self.speed_light_ambient = (1 - self.damping) * (
            self.speed_light_ambient * inertia + (1 - inertia) * step
        )
        self.light_ambient = self.light_ambient + self.speed_light_ambient
        # update mesh color
        step = -self.mesh_color_b * 0.00001
        self.speed_mesh_color = (1 - self.damping) * (
            self.speed_mesh_color * inertia + (1 - inertia) * step
        )
        self.mesh_color = self.mesh_color + self.speed_mesh_color

        self.iter += 1
        return energy, image, diff_image


class MeshRGBFitterWithPoseMultiFrame:
    """Class to fit a deformable mesh to multiple color images."""

    def __init__(
        self,
        vertices: np.ndarray,
        faces: np.ndarray,
        euler_init: np.ndarray,
        translation_init: np.ndarray,
        default_color: np.ndarray,
        default_light_directional: np.ndarray,
        default_light_ambient: float,
        cregu: float = 2000,
        cdata: float = 1,
        inertia: float = 0.97,
        damping: float = 0.15,
        update_lights: bool = True,
        update_color: bool = True,
    ):
        self.cregu = cregu
        self.cdata = cdata
        self.inertia = inertia
        self.damping = damping
        self.step_factor_vertices = 0.0005
        self.step_max_vertices = 0.5
        self.step_factor_quaternion = 0.00005
        self.step_max_quaternion = 0.05
        self.step_factor_translation = 0.00004
        self.step_max_translation = 0.1

        self.default_color = default_color
        self.default_light_directional = default_light_directional
        self.default_light_ambient = default_light_ambient
        self.update_lights = update_lights
        self.update_color = update_color
        self.mesh = ColoredTriMesh(faces, vertices, nb_colors=3)
        object_center = vertices.mean(axis=0)
        self.object_radius = np.max(np.std(vertices, axis=0))
        self.camera_center = object_center + np.array([0, 0, 6]) * self.object_radius

        self.scene = Scene3D()
        self.scene.set_mesh(self.mesh)
        self.rigid_energy = LaplacianRigidEnergy(self.mesh, vertices, cregu)
        self.vertices_init = copy.copy(vertices)
        self.Hfactorized = None
        self.Hpreconditioner = None
        self.set_mesh_transform_init(euler=euler_init, translation=translation_init)
        self.store_backward: Dict[str, Any] = {}
        self.reset()

    def set_background_color(self, background_color: np.ndarray) -> None:
        self.scene.set_background_color(background_color)

    def set_mesh_transform_init(
        self, euler: np.ndarray, translation: np.ndarray
    ) -> None:
        self.transform_quaternion_init = scipy.spatial.transform.Rotation.from_euler(
            "zyx", euler
        ).as_quat()
        self.transform_translation_init = translation

    def reset(self) -> None:
        self.vertices = copy.copy(self.vertices_init)
        self.speed_vertices = np.zeros(self.vertices.shape)
        self.transform_quaternion = copy.copy(self.transform_quaternion_init)
        self.transform_translation = copy.copy(self.transform_translation_init)
        self.speed_translation = np.zeros(3)
        self.speed_quaternion = np.zeros(4)

        self.mesh_color = copy.copy(self.default_color)
        self.light_directional = copy.copy(self.default_light_directional)
        self.light_ambient = self.default_light_ambient

        self.speed_light_directional = np.zeros(self.light_directional.shape)
        self.speed_light_ambient = 0.0
        self.speed_mesh_color = np.zeros(self.mesh_color.shape)

    def set_images(
        self, mesh_images: List[np.ndarray], focal: Optional[float] = None
    ) -> None:
        self.width = mesh_images[0].shape[1]
        self.height = mesh_images[0].shape[0]
        assert mesh_images[0].ndim == 3
        self.mesh_images = mesh_images
        if focal is None:
            focal = 2 * self.width

        rot = np.array([[1, 0, 0], [0, -1, 0], [0, 0, -1]])
        trans = -rot.T.dot(self.camera_center)
        intrinsic = np.array(
            [[focal, 0, self.width / 2], [0, focal, self.height / 2], [0, 0, 1]]
        )
        extrinsic = np.column_stack((rot, trans))
        self.camera = Camera(
            extrinsic=extrinsic,
            intrinsic=intrinsic,
            width=self.width,
            height=self.height,
        )
        self.iter = 0

    def set_image(self, mesh_image: np.ndarray, focal: Optional[float] = None) -> None:
        self.width = mesh_image.shape[1]
        self.height = mesh_image.shape[0]
        assert mesh_image.ndim == 3
        self.mesh_image = mesh_image
        if focal is None:
            focal = 2 * self.width

        rot = np.array([[1, 0, 0], [0, -1, 0], [0, 0, -1]])
        trans = -rot.T.dot(self.camera_center)
        intrinsic = np.array(
            [[focal, 0, self.width / 2], [0, focal, self.height / 2], [0, 0, 1]]
        )
        extrinsic = np.column_stack((rot, trans))
        self.camera = Camera(
            extrinsic=extrinsic,
            intrinsic=intrinsic,
            width=self.width,
            height=self.height,
        )
        self.iter = 0

    def render(self, idframe: Optional[int] = None) -> np.ndarray:
        unormalized_quaternion = self.transform_quaternion[idframe]
        q_normalized = normalize(
            unormalized_quaternion
        )  # that will lead to a gradient that is in the tangent space
        vertices_transformed = (
            qrot(q_normalized, self.vertices) + self.transform_translation[idframe]
        )
        self.mesh.set_vertices(vertices_transformed)
        self.scene.set_light(
            light_directional=self.light_directional, light_ambient=self.light_ambient
        )
        self.mesh.set_vertices_colors(
            np.tile(self.mesh_color, (self.mesh.nb_vertices, 1))
        )
        image = self.scene.render(self.camera)
        self.store_backward["render"] = (idframe, unormalized_quaternion, q_normalized)
        return image

    def clear_gradients(self) -> None:
        self.light_directional_b = np.zeros(self.light_directional.shape)
        self.light_ambient_b = 0
        self._vertices_b = np.zeros(self.vertices.shape)
        self.transform_quaternion_b = np.zeros(self.transform_quaternion.shape)
        self.transform_translation_b = np.zeros(self.transform_translation.shape)
        self.mesh_color_b = np.zeros(self.mesh_color.shape)
        self.store_backward = {}

    def render_backward(self, image_b: np.ndarray) -> None:
        assert self.mesh is not None
        assert self.scene.mesh is not None
        idframe, unormalized_quaternion, q_normalized = self.store_backward["render"]
        self.scene.clear_gradients()
        self.scene.render_backward(image_b)
        assert self.mesh.vertices_colors_b is not None  # helping mypy
        self.mesh_color_b += np.sum(self.mesh.vertices_colors_b, axis=0)
        self.light_directional_b += self.scene.light_directional_b
        self.light_ambient_b += self.scene.light_ambient_b
        vertices_transformed_b = self.scene.mesh._vertices_b
        self.transform_translation_b[idframe] += np.sum(vertices_transformed_b, axis=0)
        q_normalized_b, _vertices_b = qrot_backward(
            q_normalized, self.vertices, vertices_transformed_b
        )
        self._vertices_b += _vertices_b
        self.transform_quaternion_b[idframe] += normalize_backward(
            unormalized_quaternion, q_normalized_b
        )  # that will lead to a gradient that is in the tangent space

    def energy_data(
        self, vertices: np.ndarray
    ) -> Tuple[float, List[np.ndarray], List[np.ndarray]]:
        self.vertices = vertices
        images: List[np.ndarray] = []
        diff_images: List[np.ndarray] = []
        energy_datas: List[float] = []
        self.clear_gradients()
        coef_data = self.cdata / self.nb_frames
        for idframe in range(self.nb_frames):
            image = self.render(idframe=idframe)

            diff_image = np.sum(
                (image[idframe] - self.mesh_images[idframe]) ** 2, axis=2
            )
            images.append(image)
            diff_images.append(diff_image)

            image_b = coef_data * 2 * (image[idframe] - self.mesh_images[idframe])
            energy_data_image = coef_data * np.sum(diff_image)
            energy_datas.append(energy_data_image)
            self.render_backward(image_b)
        energy_data = float(np.sum(energy_datas))

        return energy_data, images, diff_images

    def step(
        self, check_gradient: bool = False
    ) -> Tuple[float, List[np.ndarray], List[np.ndarray]]:
        self.vertices = self.vertices - np.mean(self.vertices, axis=0)[None, :]

        self.nb_frames = len(self.mesh_images)

        energy_data, image, diff_image = self.energy_data(self.vertices)
        (
            energy_rigid,
            grad_rigidity,
            _,
        ) = self.rigid_energy.evaluate(self.vertices)

        if check_gradient:

            def func(x: np.ndarray) -> np.ndarray:
                return np.array(self.rigid_energy.evaluate(x)[0])

            check_jacobian_finite_differences(
                grad_rigidity.flatten(), func, self.vertices
            )

            def func(x: np.ndarray) -> np.ndarray:
                return np.array(self.energy_data(x)[0])

            grad_data = self._vertices_b.copy()
            check_jacobian_finite_differences(grad_data.flatten(), func, self.vertices)

        energy = energy_data + energy_rigid
        print(
            f"iter {self.iter} Energy={energy} : EData={energy_data} E_rigid={energy_rigid}"
        )

        if self.iter < 500:
            self._vertices_b = (
                self._vertices_b - np.mean(self._vertices_b, axis=0)[None, :]
            )
        # update v
        grad = self._vertices_b + grad_rigidity

        def mult_and_clamp(x: np.ndarray, a: float, t: float) -> np.ndarray:
            return np.minimum(np.maximum(x * a, -t), t)

        inertia = self.inertia

        # update vertices
        step_vertices = mult_and_clamp(
            -grad, self.step_factor_vertices, self.step_max_vertices
        )

        self.speed_vertices = (1 - self.damping) * (
            self.speed_vertices * inertia + (1 - inertia) * step_vertices
        )
        self.vertices = self.vertices + self.speed_vertices
        # update rotation

        step_quaternion = mult_and_clamp(
            -self.transform_quaternion_b,
            self.step_factor_quaternion,
            self.step_max_quaternion,
        )
        self.speed_quaternion = (1 - self.damping) * (
            self.speed_quaternion * inertia + (1 - inertia) * step_quaternion
        )
        self.transform_quaternion = self.transform_quaternion + self.speed_quaternion
        self.transform_quaternion = self.transform_quaternion / np.linalg.norm(
            self.transform_quaternion
        )
        # update translation
        step_translation = mult_and_clamp(
            -self.transform_translation_b,
            self.step_factor_translation,
            self.step_max_translation,
        )
        self.speed_translation = (1 - self.damping) * (
            self.speed_translation * inertia + (1 - inertia) * step_translation
        )
        self.transform_translation = self.transform_translation + self.speed_translation
        # update directional light
        step = -self.light_directional_b * 0.0001
        self.speed_light_directional = (1 - self.damping) * (
            self.speed_light_directional * inertia + (1 - inertia) * step
        )
        self.light_directional = self.light_directional + self.speed_light_directional
        # update ambient light
        step_light_ambient = -self.light_ambient_b * 0.0001
        self.speed_light_ambient = (1 - self.damping) * (
            self.speed_light_ambient * inertia + (1 - inertia) * step_light_ambient
        )
        self.light_ambient = self.light_ambient + self.speed_light_ambient
        # update mesh color
        step = -self.mesh_color_b * 0.00001
        self.speed_mesh_color = (1 - self.damping) * (
            self.speed_mesh_color * inertia + (1 - inertia) * step
        )
        self.mesh_color = self.mesh_color + self.speed_mesh_color

        self.iter += 1
        return energy, image, diff_image



# type: ignore
"""Modules containing pytorch classes to fit 3D meshes to images using differentiable rendering."""

import copy
from typing import Callable, Optional, Tuple

import numpy as np
import scipy.sparse.linalg
import scipy.spatial.transform.rotation
import torch

from . import CameraPytorch, LaplacianRigidEnergyPytorch, Scene3DPytorch
from .. import LaplacianRigidEnergy
from .triangulated_mesh_pytorch import ColoredTriMeshPytorch as ColoredTriMesh


def print_grad(name: str) -> Callable[[torch.Tensor], None]:
    # to visualize the gradient of a variable use
    # variable_name.register_hook(print_grad('variable_name'))
    def hook(grad: torch.Tensor) -> None:
        print(f"grad {name} = {grad}")

    return hook


def qrot(q: torch.Tensor, v: torch.Tensor) -> torch.Tensor:
    qr = q[None, :].repeat(v.shape[0], 1)
    qvec = qr[:, :-1]
    uv = torch.cross(qvec, v, dim=1)
    uuv = torch.cross(qvec, uv, dim=1)
    return v + 2 * (qr[:, [3]] * uv + uuv)


class MeshDepthFitterEnergy(torch.nn.Module):
    """Pytorch module to fit a deformable mesh to a depth image."""

    def __init__(
        self,
        vertices: np.ndarray,
        faces: np.ndarray,
        euler_init: np.ndarray,
        translation_init: np.ndarray,
        cregu: float = 2000,
    ):
        super(MeshDepthFitterEnergy, self).__init__()
        self.mesh = ColoredTriMesh(
            faces=faces[:, ::-1].copy(),
            vertices=vertices,
            colors=np.zeros((vertices.shape[0], 0)),
        )  # we do a copy to avoid negative stride not supported by pytorch
        object_center = vertices.mean(axis=0)
        object_radius = np.max(np.std(vertices, axis=0))
        self.camera_center = object_center + np.array([-0.5, 0, 5]) * object_radius
        self.scene = Scene3DPytorch()
        self.scene.set_mesh(self.mesh)
        self.rigid_energy = LaplacianRigidEnergyPytorch(self.mesh, vertices, cregu)
        self.Vinit = copy.copy(self.mesh.vertices)
        self.Hfactorized = None
        self.Hpreconditioner = None
        self.transform_quaternion_init = scipy.spatial.transform.Rotation.from_euler(
            "zyx", euler_init
        ).as_quat()
        self.transform_translation_init = translation_init
        self._vertices = torch.nn.Parameter(
            torch.tensor(self.Vinit, dtype=torch.float64)
        )
        self.quaternion = torch.nn.Parameter(
            torch.tensor(self.transform_quaternion_init, dtype=torch.float64)
        )
        self.translation = torch.nn.Parameter(
            torch.tensor(self.transform_translation_init, dtype=torch.float64)
        )

    def set_max_depth(self, max_depth: float) -> None:
        self.max_depth = max_depth
        self.scene.set_background_color(np.array([max_depth], dtype=np.float64))

    def set_depth_scale(self, depth_scale: float) -> None:
        self.depthScale = depth_scale

    def set_image(
        self,
        mesh_image: np.ndarray,
        focal: Optional[float] = None,
        distortion: Optional[np.ndarray] = None,
    ) -> None:
        self.width = mesh_image.shape[1]
        self.height = mesh_image.shape[0]
        assert mesh_image.ndim == 2
        self.mesh_image = mesh_image
        if focal is None:
            focal = 2 * self.width

        rot = np.array([[1, 0, 0], [0, -1, 0], [0, 0, -1]])
        t = -rot.T.dot(self.camera_center)
        intrinsic = np.array(
            [[focal, 0, self.width / 2], [0, focal, self.height / 2], [0, 0, 1]]
        )
        extrinsic = np.column_stack((rot, t))
        self.camera = CameraPytorch(
            extrinsic=extrinsic,
            intrinsic=intrinsic,
            distortion=distortion,
            height=mesh_image.shape[0],
            width=mesh_image.shape[1],
        )
        self.iter = 0

    def forward(self) -> torch.Tensor:
        q_normalized = self.quaternion / self.quaternion.norm()
        print(self.quaternion.norm())
        vertices_centered = self._vertices - torch.mean(self._vertices, dim=0)[None, :]
        v_transformed = qrot(q_normalized, vertices_centered) + self.translation
        self.mesh.set_vertices(v_transformed)
        depth_scale = 1 * self.depthScale
        depth = self.scene.render_depth(
            self.camera,
            depth_scale=depth_scale,
        )
        depth = torch.clamp(depth, 0, self.max_depth)
        diff_image = torch.sum(
            (depth - torch.tensor(self.mesh_image[:, :, None])) ** 2, dim=2
        )
        self.depth = depth
        self.diff_image = diff_image
        energy_data = torch.sum(diff_image)
        energy_rigid = self.rigid_energy.evaluate(
            self._vertices,
        )[0]
        energy = energy_data + energy_rigid
        self.loss = energy_data + energy_rigid
        print("Energy=%f : EData=%f E_rigid=%f" % (energy, energy_data, energy_rigid))
        return self.loss


class MeshDepthFitterPytorchOptim:
    """Pytorch optimizer to fit a deformable mesh to an image."""

    def __init__(
        self,
        vertices: np.ndarray,
        faces: np.ndarray,
        euler_init: np.ndarray,
        translation_init: np.ndarray,
        cregu: float = 2000,
        lr: float = 0.8,
    ):
        self.energy = MeshDepthFitterEnergy(
            vertices, faces, euler_init, translation_init, cregu
        )
        params = self.energy.parameters()
        self.optimizer = torch.optim.LBFGS(params, lr=0.8, max_iter=1)
        # self.optimizer = torch.optim.SGD(params, lr=0.000005, momentum=0.1,
        # dampening=0.1        )
        # self.optimizer =torch.optim.RMSprop(params, lr=1e-3, alpha=0.99,  eps=1e-8,
        # weight_decay=0,  momentum=0.001)
        # self.optimizer = torch.optim.Adadelta(params, lr=0.1, rho=0.95,
        #   eps=1e-6,  weight_decay=0)
        # self.optimizer = torch.optim.Adagrad(self.energy.parameters(), lr=0.02)

    def set_image(self, depth_image: np.ndarray, focal: float) -> None:
        self.energy.set_image(depth_image, focal=focal)

    def set_max_depth(self, max_depth: float) -> None:
        self.energy.set_max_depth(max_depth)

    def set_depth_scale(self, depth_scale: float) -> None:
        self.energy.set_depth_scale(depth_scale)

    def step(self) -> Tuple[torch.Tensor, np.ndarray, np.ndarray]:
        def closure() -> float:
            self.optimizer.zero_grad()
            loss = self.energy()
            loss.backward()
            return loss

        self.optimizer.step(closure)
        # self.iter += 1

        return (
            self.energy.loss,
            self.energy.depth[:, :, 0].detach().numpy(),
            self.energy.diff_image.detach().numpy(),
        )


def mult_and_clamp(x: np.ndarray, a: float, t: float) -> np.ndarray:
    return np.minimum(np.maximum(x * a, -t), t)


class MeshDepthFitter:
    """Class to fit a deformable mesh to a depth image."""

    def __init__(
        self,
        vertices: np.ndarray,
        faces: np.ndarray,
        euler_init: np.ndarray,
        translation_init: np.ndarray,
        cregu: float = 2000,
        inertia: float = 0.96,
        damping: float = 0.05,
    ):
        self.cregu = cregu
        self.inertia = inertia
        self.damping = damping
        self.step_factor_vertices = 0.0005
        self.step_max_vertices = 0.5
        self.step_factor_quaternion = 0.00006
        self.step_max_quaternion = 0.1
        self.step_factor_translation = 0.00005
        self.step_max_translation = 0.1

        self.mesh = ColoredTriMesh(
            faces=faces.copy(),
            vertices=vertices,
            colors=np.zeros((vertices.shape[0], 0)),
        )  # we do a copy to avoid negative stride not support by pytorch
        object_center = vertices.mean(axis=0) + translation_init
        object_radius = np.max(np.std(vertices, axis=0))
        self.camera_center = object_center + np.array([-0.5, 0, 5]) * object_radius

        self.scene = Scene3DPytorch()
        self.scene.set_mesh(self.mesh)
        self.rigid_energy = LaplacianRigidEnergy(self.mesh, vertices, cregu)
        self.vertices_init = torch.tensor(copy.copy(vertices))
        self.Hfactorized = None
        self.Hpreconditioner = None
        self.set_mesh_transform_init(euler=euler_init, translation=translation_init)
        self.reset()

    def set_mesh_transform_init(
        self, euler: np.ndarray, translation: np.ndarray
    ) -> None:
        self.transform_quaternion_init = scipy.spatial.transform.Rotation.from_euler(
            "zyx", euler
        ).as_quat()
        self.transform_translation_init = translation

    def reset(self) -> None:
        self.vertices = copy.copy(self.vertices_init)
        self.speed_vertices = np.zeros(self.vertices_init.shape)
        self.transform_quaternion = copy.copy(self.transform_quaternion_init)
        self.transform_translation = copy.copy(self.transform_translation_init)
        self.speed_translation = np.zeros(3)
        self.speed_quaternion = np.zeros(4)

    def set_max_depth(self, max_depth: float) -> None:
        self.max_depth = max_depth
        self.scene.set_background_color([max_depth])

    def set_depth_scale(self, depth_scale: float) -> None:
        self.depthScale = depth_scale

    def set_image(
        self,
        mesh_image: np.ndarray,
        focal: Optional[float] = None,
        distortion: Optional[np.ndarray] = None,
    ) -> None:
        self.width = mesh_image.shape[1]
        self.height = mesh_image.shape[0]
        assert mesh_image.ndim == 2
        self.mesh_image = mesh_image
        if focal is None:
            focal = 2 * self.width

        rot = np.array([[1, 0, 0], [0, -1, 0], [0, 0, -1]])
        trans = -rot.T.dot(self.camera_center)
        intrinsic = np.array(
            [[focal, 0, self.width / 2], [0, focal, self.height / 2], [0, 0, 1]]
        )
        extrinsic = np.column_stack((rot, trans))
        self.camera = CameraPytorch(
            extrinsic=extrinsic,
            intrinsic=intrinsic,
            width=self.width,
            height=self.height,
            distortion=distortion,
        )
        self.iter = 0

    def step(self) -> Tuple[float, np.ndarray, np.ndarray]:
        self.vertices = self.vertices - torch.mean(self.vertices, dim=0)[None, :]
        # vertices_with_grad = self.vertices.clone().requires_grad(True)
        vertices_with_grad = self.vertices.clone().detach().requires_grad_(True)
        vertices_with_grad_centered = (
            vertices_with_grad - torch.mean(vertices_with_grad, dim=0)[None, :]
        )
        quaternion_with_grad = torch.tensor(
            self.transform_quaternion, dtype=torch.float64, requires_grad=True
        )
        translation_with_grad = torch.tensor(
            self.transform_translation, dtype=torch.float64, requires_grad=True
        )

        q_normalized = (
            quaternion_with_grad / quaternion_with_grad.norm()
        )  # that will lead to a gradient that is in the tangent space
        vertices_with_grad_transformed = (
            qrot(q_normalized, vertices_with_grad_centered) + translation_with_grad
        )

        self.mesh.set_vertices(vertices_with_grad_transformed)

        depth_scale = 1 * self.depthScale
        depth = self.scene.render_depth(self.camera, depth_scale=depth_scale)
        depth = torch.clamp(depth, 0, self.max_depth)

        diff_image = torch.sum(
            (depth - torch.tensor(self.mesh_image[:, :, None])) ** 2, dim=2
        )
        loss = torch.sum(diff_image)

        loss.backward()
        energy_data = loss.detach().numpy()

        grad_data = vertices_with_grad.grad.numpy()

        (
            energy_rigid,
            grad_rigidity,
            _,
        ) = self.rigid_energy.evaluate(self.vertices.numpy())
        energy = energy_data + energy_rigid
        print("Energy=%f : EData=%f E_rigid=%f" % (energy, energy_data, energy_rigid))

        # update v
        grad = grad_data + grad_rigidity

        # update vertices
        step_vertices = mult_and_clamp(
            -grad, self.step_factor_vertices, self.step_max_vertices
        )
        self.speed_vertices = (1 - self.damping) * (
            self.speed_vertices * self.inertia + (1 - self.inertia) * step_vertices
        )
        self.vertices = self.vertices + torch.tensor(self.speed_vertices)
        # update rotation
        step_quaternion = mult_and_clamp(
            -quaternion_with_grad.grad.numpy(),
            self.step_factor_quaternion,
            self.step_max_quaternion,
        )
        self.speed_quaternion = (1 - self.damping) * (
            self.speed_quaternion * self.inertia + (1 - self.inertia) * step_quaternion
        )
        self.transform_quaternion = self.transform_quaternion + self.speed_quaternion
        self.transform_quaternion = self.transform_quaternion / np.linalg.norm(
            self.transform_quaternion
        )
        # update translation

        step_translation = mult_and_clamp(
            -translation_with_grad.grad.numpy(),
            self.step_factor_translation,
            self.step_max_translation,
        )
        self.speed_translation = (1 - self.damping) * (
            self.speed_translation * self.inertia
            + (1 - self.inertia) * step_translation
        )
        self.transform_translation = self.transform_translation + self.speed_translation

        self.iter += 1
        return energy, depth[:, :, 0].detach().numpy(), diff_image.detach().numpy()


class MeshRGBFitterWithPose:
    """Class to fit a deformable mesh to a color image."""

    def __init__(
        self,
        vertices: np.ndarray,
        faces: np.ndarray,
        euler_init: np.ndarray,
        translation_init: np.ndarray,
        default_color: np.ndarray,
        default_light_directional: np.ndarray,
        default_light_ambient: float,
        cregu: float = 2000,
        inertia: float = 0.96,
        damping: float = 0.05,
        update_lights: bool = True,
        update_color: bool = True,
    ):
        self.cregu = cregu

        self.inertia = inertia
        self.damping = damping
        self.step_factor_vertices = 0.0005
        self.step_max_vertices = 0.5
        self.step_factor_quaternion = 0.00006
        self.step_max_quaternion = 0.05
        self.step_factor_translation = 0.00005
        self.step_max_translation = 0.1

        self.default_color = default_color
        self.default_light_directional = default_light_directional
        self.default_light_ambient = default_light_ambient
        self.update_lights = update_lights
        self.update_color = update_color
        self.mesh = ColoredTriMesh(
            faces=faces.copy(),
            vertices=vertices,
            colors=np.zeros((vertices.shape[0], 0)),
        )  # we do a copy to avoid negative stride not support by pytorch
        object_center = vertices.mean(axis=0) + translation_init
        object_radius = np.max(np.std(vertices, axis=0))
        self.camera_center = object_center + np.array([0, 0, 9]) * object_radius

        self.scene = Scene3DPytorch()
        self.scene.set_mesh(self.mesh)
        self.rigid_energy = LaplacianRigidEnergyPytorch(self.mesh, vertices, cregu)
        self.vertices_init = torch.tensor(copy.copy(vertices))
        self.Hfactorized = None
        self.Hpreconditioner = None
        self.set_mesh_transform_init(euler=euler_init, translation=translation_init)
        self.reset()

    def set_background_color(self, background_color: np.ndarray) -> None:
        self.scene.set_background_color(background_color)

    def set_mesh_transform_init(
        self, euler: np.ndarray, translation: np.ndarray
    ) -> None:
        self.transform_quaternion_init = scipy.spatial.transform.Rotation.from_euler(
            "zyx", euler
        ).as_quat()
        self.transform_translation_init = translation

    def reset(self) -> None:
        self.vertices = copy.copy(self.vertices_init)
        self.speed_vertices = np.zeros(self.vertices.shape)
        self.transform_quaternion = copy.copy(self.transform_quaternion_init)
        self.transform_translation = copy.copy(self.transform_translation_init)
        self.speed_translation = np.zeros(3)
        self.speed_quaternion = np.zeros(4)

        self.mesh_color = copy.copy(self.default_color)
        self.light_directional = copy.copy(self.default_light_directional)
        self.light_ambient = copy.copy(self.default_light_ambient)

        self.speed_light_directional = np.zeros(self.light_directional.shape)
        self.speed_light_ambient = 0.0
        self.speed_mesh_color = np.zeros(self.mesh_color.shape)

    def set_image(
        self,
        mesh_image: np.ndarray,
        focal: Optional[float] = None,
        distortion: Optional[np.ndarray] = None,
    ) -> None:
        self.width = mesh_image.shape[1]
        self.height = mesh_image.shape[0]
        assert mesh_image.ndim == 3
        self.mesh_image = mesh_image
        if focal is None:
            focal = 2 * self.width

        rot = np.array([[1, 0, 0], [0, -1, 0], [0, 0, -1]])
        trans = -rot.T.dot(self.camera_center)
        intrinsic = np.array(
            [[focal, 0, self.width / 2], [0, focal, self.height / 2], [0, 0, 1]]
        )
        extrinsic = np.column_stack((rot, trans))
        self.camera = CameraPytorch(
            extrinsic=extrinsic,
            intrinsic=intrinsic,
            width=self.width,
            height=self.height,
            distortion=distortion,
        )
        self.iter = 0

    def step(self) -> Tuple[float, np.ndarray, np.ndarray]:
        self.vertices = self.vertices - torch.mean(self.vertices, dim=0)[None, :]
        vertices_with_grad = self.vertices.clone().detach().requires_grad_(True)
        vertices_with_grad_centered = (
            vertices_with_grad - torch.mean(vertices_with_grad, dim=0)[None, :]
        )
        quaternion_with_grad = torch.tensor(
            self.transform_quaternion, dtype=torch.float64, requires_grad=True
        )
        translation_with_grad = torch.tensor(
            self.transform_translation, dtype=torch.float64, requires_grad=True
        )

        light_directional_with_grad = torch.tensor(
            self.light_directional, dtype=torch.float64, requires_grad=True
        )
        light_ambient_with_grad = torch.tensor(
            self.light_ambient, dtype=torch.float64, requires_grad=True
        )
        mesh_color_with_grad = torch.tensor(
            self.mesh_color, dtype=torch.float64, requires_grad=True
        )

        q_normalized = (
            quaternion_with_grad / quaternion_with_grad.norm()
        )  # that will lead to a gradient that is in the tangent space
        vertices_with_grad_transformed = (
            qrot(q_normalized, vertices_with_grad_centered) + translation_with_grad
        )
        self.mesh.set_vertices(vertices_with_grad_transformed)

        self.scene.set_light(
            light_directional=light_directional_with_grad,
            light_ambient=light_ambient_with_grad,
        )
        self.mesh.set_vertices_colors(
            mesh_color_with_grad.repeat([self.mesh.nb_vertices, 1])
        )

        image = self.scene.render(self.camera)

        diff_image = torch.sum((image - torch.tensor(self.mesh_image)) ** 2, dim=2)
        loss = torch.sum(diff_image)

        loss.backward()
        energy_data = loss.detach().numpy()

        grad_data = vertices_with_grad.grad

        (
            energy_rigid,
            grad_rigidity,
            approx_hessian_rigidity,
        ) = self.rigid_energy.evaluate(self.vertices)
        energy = energy_data + energy_rigid.numpy()
        print("Energy=%f : EData=%f E_rigid=%f" % (energy, energy_data, energy_rigid))

        # update v
        grad = grad_data + grad_rigidity

        inertia = self.inertia

        # update vertices
        step_vertices = mult_and_clamp(
            -grad.numpy(), self.step_factor_vertices, self.step_max_vertices
        )
        self.speed_vertices = (1 - self.damping) * (
            self.speed_vertices * inertia + (1 - inertia) * step_vertices
        )
        self.vertices = self.vertices + torch.tensor(self.speed_vertices)
        # update rotation
        step_quaternion = mult_and_clamp(
            -quaternion_with_grad.grad.numpy(),
            self.step_factor_quaternion,
            self.step_max_quaternion,
        )
        self.speed_quaternion = (1 - self.damping) * (
            self.speed_quaternion * inertia + (1 - inertia) * step_quaternion
        )
        self.transform_quaternion = self.transform_quaternion + self.speed_quaternion
        self.transform_quaternion = self.transform_quaternion / np.linalg.norm(
            self.transform_quaternion
        )

        # update translation
        step_translation = mult_and_clamp(
            -translation_with_grad.grad.numpy(),
            self.step_factor_translation,
            self.step_max_translation,
        )
        self.speed_translation = (1 - self.damping) * (
            self.speed_translation * inertia + (1 - inertia) * step_translation
        )
        self.transform_translation = self.transform_translation + self.speed_translation
        # update directional light
        step = -light_directional_with_grad.grad.numpy() * 0.0001
        self.speed_light_directional = (1 - self.damping) * (
            self.speed_light_directional * inertia + (1 - inertia) * step
        )
        self.light_directional = self.light_directional + self.speed_light_directional
        # update ambient light
        step = -light_ambient_with_grad.grad.numpy() * 0.0001
        self.speed_light_ambient = (1 - self.damping) * (
            self.speed_light_ambient * inertia + (1 - inertia) * step
        )
        self.light_ambient = self.light_ambient + self.speed_light_ambient
        # update mesh color
        step = -mesh_color_with_grad.grad.numpy() * 0.00001
        self.speed_mesh_color = (1 - self.damping) * (
            self.speed_mesh_color * inertia + (1 - inertia) * step
        )
        self.mesh_color = self.mesh_color + self.speed_mesh_color

        self.iter += 1
        return energy, image.detach().numpy(), diff_image.detach().numpy()



#! /usr/bin/env python
"""Function to oad wavefront OBJ files

See http://www.fileformat.info/format/wavefrontobj/.
At the moment only v and f keywords are supported
"""
from typing import List, Tuple

import numpy as np


def read_obj(filename: str) -> Tuple[np.ndarray, np.ndarray]:
    """Very limited obj file loading.

    At the moment only v and f keywords are supported
    """

    faces = []
    vertices = []
    fid = open(filename, "r")
    node_counter = 0
    while True:
        line = fid.readline()
        if line == "":
            break
        while line.endswith("\\"):
            # Remove backslash and concatenate with next line
            line = line[:-1] + fid.readline()
        if line.startswith("v"):
            coord = line.split()
            coord.pop(0)
            node_counter += 1
            vertices.append(np.array([float(c) for c in coord]))

        elif line.startswith("f "):
            fields = line.split()
            fields.pop(0)

            # in some obj faces are defined as -70//-70 -69//-69 -62//-62
            cleaned_fields: List[int] = []
            for f in fields:
                v = int(f.split("/")[0]) - 1
                if v < 0:
                    v = node_counter + v
                cleaned_fields.append(v)
            faces.append(np.array(cleaned_fields))

    faces_np = np.row_stack(faces)
    vertices_np = np.row_stack(vertices)
    return faces_np, vertices_np


def save_obj(filename: str, vertices: np.ndarray, faces: np.ndarray) -> None:
    with open(filename, "w") as f:
        for vertex in vertices:
            f.write(f"v {vertex[0]:08f} {vertex[1]:08f} {vertex[2]:08f}\n")
        for face in faces:
            f.write(f"f {face[0]+1} {face[1]+1} {face[2]+1}\n")



"""Tensorflow interface to deodr."""

from typing import Callable, Iterable, Optional, Tuple

import numpy as np
import tensorflow as tf

from .. import differentiable_renderer_cython  # type: ignore
from ..differentiable_renderer import Camera, Scene3D


class CameraTensorflow(Camera):
    """Tensorflow implementation of the camera class."""

    def __init__(
        self,
        extrinsic: np.ndarray,
        intrinsic: np.ndarray,
        height: int,
        width: int,
        distortion: Optional[np.ndarray] = None,
    ):
        super().__init__(
            extrinsic, intrinsic, height, width, distortion=distortion, checks=False
        )

    def world_to_camera(self, points_3d: tf.Tensor) -> tf.Tensor:
        assert isinstance(points_3d, tf.Tensor)
        return tf.linalg.matmul(
            tf.concat(
                (points_3d, tf.ones((points_3d.shape[0], 1), dtype=points_3d.dtype)),
                axis=1,
            ),
            tf.constant(self.extrinsic.T),
        )

    def left_mul_intrinsic(self, projected: tf.Tensor) -> tf.Tensor:
        assert isinstance(projected, tf.Tensor)
        return tf.linalg.matmul(
            tf.concat(
                (projected, tf.ones((projected.shape[0], 1), dtype=projected.dtype)),
                axis=1,
            ),
            tf.constant(self.intrinsic[:2, :].T),
        )

    def column_stack(self, values: Iterable[tf.Tensor]) -> tf.Tensor:
        return tf.stack(values, axis=1)


def TensorflowDifferentiableRender2D(
    ij: tf.Tensor, colors: tf.Tensor, scene: "Scene3DTensorflow"
) -> Tuple[tf.Tensor, Callable[[tf.Tensor], Tuple[tf.Tensor, tf.Tensor]]]:
    """Tensorflow implementation of the 2D rendering function."""

    @tf.custom_gradient
    def forward(
        ij: tf.Tensor, colors: tf.Tensor
    ) -> Tuple[tf.Tensor, Callable[[tf.Tensor], Tuple[tf.Tensor, tf.Tensor]]]:
        # using inner function as we don't differentiate w.r.t scene
        nb_color_channels = colors.shape[1]
        image = np.empty(
            (scene.scene_2d.height, scene.scene_2d.width, nb_color_channels)
        )
        z_buffer = np.empty((scene.scene_2d.height, scene.scene_2d.width))
        scene.scene_2d.ij = np.array(ij)  # should automatically detached according to
        # https://pytorch.org/docs/master/notes/extending.html
        scene.scene_2d.colors = np.array(colors)

        scene.scene_2d.depths = np.array(scene.scene_2d.depths)
        differentiable_renderer_cython.renderSceneCpp(
            scene.scene_2d, 1, image, z_buffer
        )

        def backward(image_b: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor]:
            assert scene.scene_2d.colors is not None
            scene.scene_2d.uv_b = np.zeros(scene.scene_2d.uv.shape)
            scene.scene_2d.ij_b = np.zeros(scene.scene_2d.ij.shape)
            scene.scene_2d.shade_b = np.zeros(scene.scene_2d.shade.shape)
            scene.scene_2d.colors_b = np.zeros(scene.scene_2d.colors.shape)
            scene.scene_2d.texture_b = np.zeros(scene.scene_2d.texture.shape)
            image_copy = (
                image.copy()
            )  # making a copy to avoid removing antialiasing on the image returned by
            # the forward pass (the c++ back-propagation undoes antialiasing), could be
            # optional if we don't care about getting aliased images
            differentiable_renderer_cython.renderSceneBCpp(
                scene.scene_2d, 1, image_copy, z_buffer, image_b.numpy()
            )
            return tf.constant(scene.scene_2d.ij_b), tf.constant(
                scene.scene_2d.colors_b
            )

        return tf.convert_to_tensor(image), backward

    return forward(ij, colors)


class Scene3DTensorflow(Scene3D):
    """Tensorflow implementation of deodr 3D scenes."""

    def __init__(self) -> None:
        super().__init__()

    def set_light(self, light_directional: tf.Tensor, light_ambient: tf.Tensor) -> None:
        if not (isinstance(light_directional, tf.Tensor)):
            light_directional = tf.constant(light_directional)
        self.light_directional = light_directional
        self.light_ambient = light_ambient

    def _compute_vertices_colors_with_illumination(self) -> tf.Tensor:
        assert self.mesh is not None, "You need to provide a mesh first."
        vertices_luminosity = (
            tf.nn.relu(
                -tf.reduce_sum(
                    self.mesh.vertex_normals * self.light_directional, axis=1
                )
            )
            + self.light_ambient
        )
        return self.mesh.vertices_colors * vertices_luminosity[:, None]

    def _render_2d(self) -> tf.Tensor:
        return (
            TensorflowDifferentiableRender2D(
                self.scene_2d.ij, self.scene_2d.colors, self
            ),
            self.scene_2d.depths,
        )



"""Tensorflow interface for deodr."""

__all__ = [
    "ColoredTriMeshTensorflow",
    "Scene3DTensorflow",
    "CameraTensorflow",
    "LaplacianRigidEnergyTensorflow",
    "MeshRGBFitterWithPose",
    "MeshDepthFitter",
]

from .differentiable_renderer_tensorflow import CameraTensorflow, Scene3DTensorflow
from .laplacian_rigid_energy_tensorflow import LaplacianRigidEnergyTensorflow
from .mesh_fitter_tensorflow import MeshDepthFitter, MeshRGBFitterWithPose
from .triangulated_mesh_tensorflow import ColoredTriMeshTensorflow



"""Tensorflow implementation of a triangulated mesh."""
from typing import Optional

import numpy as np
import tensorflow as tf

from ..triangulated_mesh import ColoredTriMesh, TriMeshAdjacencies
from .tools import scipy_sparse_matrix_to_tensorflow


class TriMeshAdjacenciesTensorflow(TriMeshAdjacencies):
    """Class that stores adjacency matrices and methods that use this adjacencies using tensorflow sparse matrices.
    Unlike the TriMesh class there are no vertices stored in this class.
    """

    def __init__(self, faces: np.ndarray):
        super().__init__(faces)
        self.faces_tf = tf.constant(faces)
        self._vertices_faces_tf = scipy_sparse_matrix_to_tensorflow(
            self._vertices_faces
        )

    def compute_face_normals(self, vertices: tf.Tensor) -> tf.Tensor:
        tris = tf.gather(vertices, self.faces)
        u = tris[::, 1] - tris[::, 0]
        v = tris[::, 2] - tris[::, 0]
        n = -tf.linalg.cross(u, v) if self.clockwise else tf.linalg.cross(u, v)
        norm = tf.sqrt(tf.reduce_sum(n**2, axis=1))
        return n / norm[:, None]

    def compute_vertex_normals(self, face_normals: tf.Tensor) -> tf.Tensor:
        n = tf.sparse.sparse_dense_matmul(self._vertices_faces_tf, face_normals)
        norm = tf.sqrt(tf.reduce_sum(n**2, axis=1))
        return n / norm[:, None]

    def edge_on_silhouette(self, vertices_2d: tf.Tensor) -> np.ndarray:
        return super().edge_on_silhouette(vertices_2d.numpy())


class ColoredTriMeshTensorflow(ColoredTriMesh):
    """Tensorflow implementation of a colored triangulated mesh."""

    def __init__(
        self,
        faces: np.ndarray,
        vertices: np.ndarray,
        clockwise: bool = False,
        faces_uv: Optional[np.ndarray] = None,
        uv: Optional[np.ndarray] = None,
        texture: Optional[np.ndarray] = None,
        colors: Optional[np.ndarray] = None,
    ):
        super(ColoredTriMeshTensorflow, self).__init__(
            faces,
            vertices=vertices,
            clockwise=clockwise,
            faces_uv=faces_uv,
            uv=uv,
            texture=texture,
            colors=colors,
        )

    def compute_adjacencies(self) -> None:
        self._adjacencies = TriMeshAdjacenciesTensorflow(self.faces)

    def set_vertices_colors(self, colors: tf.Tensor) -> None:
        self.vertices_colors = colors



# type: ignore
"""Pytorch implementation of an as-rigid-as-possible energy based on the difference of laplacian with a reference shape."""

from typing import Tuple

import numpy as np
from scipy.sparse import spmatrix
import torch
from torch.sparse import DoubleTensor  # type: ignore

from ..laplacian_rigid_energy import LaplacianRigidEnergy
from .triangulated_mesh_pytorch import ColoredTriMeshPytorch


def scipy_sparse_to_torch(sparse_matrix: spmatrix) -> DoubleTensor:
    coo = sparse_matrix.tocoo()
    values = coo.data
    indices = np.vstack((coo.row, coo.col))
    i = torch.LongTensor(indices.astype(np.int32))
    v = torch.DoubleTensor(values)
    shape = coo.shape
    return DoubleTensor(i, v, torch.Size(shape))


class LaplacianRigidEnergyPytorch:
    """Pytorch class that implements an as-rigid-as-possible energy based on the difference of laplacian with a reference shape."""

    def __init__(self, mesh: ColoredTriMeshPytorch, vertices: np.ndarray, cregu: float):
        self.numpy_imp = LaplacianRigidEnergy(mesh, vertices, cregu)
        self.cT_torch = scipy_sparse_to_torch(self.numpy_imp.cT)
        self.approx_hessian = scipy_sparse_to_torch(self.numpy_imp.approx_hessian)

    def evaluate(
        self,
        vertices: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        assert isinstance(vertices, torch.Tensor)
        if vertices.requires_grad:
            diff = (vertices - self.numpy_imp.vertices_ref).flatten()
            grad_vertices = self.numpy_imp.cregu * (
                self.cT_torch.matmul(diff[:, None])
            ).reshape_as(vertices)
        else:
            diff = (vertices - torch.tensor(self.numpy_imp.vertices_ref)).flatten()
            # gradV = self.cregu*(self.cT_torch.matmul(diff[:,None])).reshape_as(V)
            # 40x slower than scipy !
            grad_vertices = torch.tensor(
                self.numpy_imp.cregu
                * (self.numpy_imp.cT * (diff[:, None].numpy())).reshape(vertices.shape)
            )

        energy = 0.5 * diff.dot(grad_vertices.flatten())
        return energy, grad_vertices, self.numpy_imp.approx_hessian
